from __future__ import annotations

import time

from numpy import dot
from numpy.linalg import norm
from openai import OpenAI

from palimpzest.constants import (
    MODEL_CARDS,
    NAIVE_EST_NUM_OUTPUT_TOKENS,
    Model,
)
from palimpzest.core.data.dataclasses import GenerationStats, OperatorCostEstimates
from palimpzest.core.elements.records import DataRecord
from palimpzest.core.lib.fields import Field, StringField
from palimpzest.query.operators.convert import LLMConvert


class RAGConvert(LLMConvert):
    def __init__(self, num_chunks_per_field: int, chunk_size: int = 1000, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # NOTE: in the future, we should abstract the embedding model to allow for different models
        self.client = None
        self.embedding_model = Model.TEXT_EMBEDDING_3_SMALL
        self.num_chunks_per_field = num_chunks_per_field
        self.chunk_size = chunk_size

        # crude adjustment factor for naive estimation in no-sentinel setting
        self.naive_quality_adjustment = 0.6

    def __str__(self):
        op = super().__str__()
        op += f"    Number of Chunks: {str(self.num_chunks_per_field)}\n"
        op += f"    Chunk Size: {str(self.chunk_size)}\n"
        return op

    def get_id_params(self):
        id_params = super().get_id_params()
        id_params = {"num_chunks_per_field": self.num_chunks_per_field, "chunk_size": self.chunk_size, **id_params}

        return id_params

    def get_op_params(self):
        op_params = super().get_op_params()
        return {"num_chunks_per_field": self.num_chunks_per_field, "chunk_size": self.chunk_size, **op_params}

    def naive_cost_estimates(self, source_op_cost_estimates: OperatorCostEstimates) -> OperatorCostEstimates:
        """
        Update the cost per record and quality estimates produced by LLMConvert's naive estimates.
        We adjust the cost per record to account for the reduced number of input tokens following
        the retrieval of relevant chunks, and we make a crude estimate of the quality degradation
        that results from using a downsized input (although this may in fact improve quality in
        some cases).
        """
        # get naive cost estimates from LLMConvert
        naive_op_cost_estimates = super().naive_cost_estimates(source_op_cost_estimates)

        # re-compute cost per record assuming we use fewer input tokens; naively assume a single input field
        est_num_input_tokens = self.num_chunks_per_field * self.chunk_size
        est_num_output_tokens = NAIVE_EST_NUM_OUTPUT_TOKENS
        model_conversion_usd_per_record = (
            MODEL_CARDS[self.model.value]["usd_per_input_token"] * est_num_input_tokens
            + MODEL_CARDS[self.model.value]["usd_per_output_token"] * est_num_output_tokens
        )

        # set refined estimate of cost per record
        naive_op_cost_estimates.cost_per_record = model_conversion_usd_per_record
        naive_op_cost_estimates.cost_per_record_lower_bound = naive_op_cost_estimates.cost_per_record
        naive_op_cost_estimates.cost_per_record_upper_bound = naive_op_cost_estimates.cost_per_record
        naive_op_cost_estimates.quality = (naive_op_cost_estimates.quality) * self.naive_quality_adjustment
        naive_op_cost_estimates.quality_lower_bound = naive_op_cost_estimates.quality
        naive_op_cost_estimates.quality_upper_bound = naive_op_cost_estimates.quality

        return naive_op_cost_estimates

    def is_image_conversion(self) -> bool:
        """RAGConvert is currently disallowed on image conversions, so this must be False."""
        return False

    def chunk_text(self, text: str, chunk_size: int) -> list[str]:
        """
        Given a text string, chunk it into substrings of length chunk_size.
        """
        chunks = []
        idx = 0
        while idx + chunk_size < len(text):
            chunks.append(text[idx : idx + chunk_size])
            idx += chunk_size
        
        if idx < len(text):
            chunks.append(text[idx:])

        return chunks

    def compute_embedding(self, text: str) -> tuple[list[float], GenerationStats]:
        """
        Compute the embedding for a text string. Return the embedding and the GenerationStats object
        that captures the cost of the operation.
        """
        # get the embedding model name
        model_name = self.embedding_model.value

        # compute the embedding
        start_time = time.time()
        response = self.client.embeddings.create(input=text, model=model_name)
        total_time = time.time() - start_time

        # extract the embedding
        embedding = response.data[0].embedding

        # compute the generation stats object
        model_card = MODEL_CARDS[model_name]
        total_input_tokens = response.usage.total_tokens
        total_input_cost = model_card["usd_per_input_token"] * total_input_tokens
        embed_stats = GenerationStats(
            model_name=model_name,  # NOTE: this should be overwritten by generation model in convert()
            total_input_tokens=total_input_tokens,
            total_output_tokens=0.0,
            total_input_cost=total_input_cost,
            total_output_cost=0.0,
            cost_per_record=total_input_cost,
            llm_call_duration_secs=total_time,
            total_llm_calls=1,
            total_embedding_llm_calls=1,
        )

        return embedding, embed_stats

    def compute_similarity(self, query_embedding: list[float], chunk_embedding: list[float]) -> float:
        """
        Compute the similarity between the query and chunk embeddings.
        """
        return dot(query_embedding, chunk_embedding) / (norm(query_embedding) * norm(chunk_embedding))

    def get_chunked_candidate(self, candidate: DataRecord, input_fields: list[str], output_fields: list[str]) -> tuple[DataRecord, GenerationStats]:
        """
        For each text field, chunk the content and compute the chunk embeddings. Then select the top-k chunks
        for each field. If a field is smaller than the chunk size, simply include the full field.
        """
        # initialize stats for embedding costs
        embed_stats = GenerationStats()

        # compute embedding for output fields
        output_fields_desc = ""
        field_desc_map = self.output_schema.field_desc_map()
        for field_name in output_fields:
            output_fields_desc += f"- {field_name}: {field_desc_map[field_name]}\n"
        query_embedding, query_embed_stats = self.compute_embedding(output_fields_desc)

        # add cost of embedding the query to embed_stats
        embed_stats += query_embed_stats

        # for each input field, chunk its content and compute the (per-chunk) embeddings
        for field_name in input_fields:
            field = candidate.get_field_type(field_name)

            # skip this field if it is not a string or a list of strings
            is_string_field = isinstance(field, StringField)
            is_list_string_field = hasattr(field, "element_type") and isinstance(field.element_type, StringField)
            if not (is_string_field or is_list_string_field):
                continue

            # if this is a list of strings, join the strings
            if is_list_string_field:
                candidate[field_name] = "[" + ", ".join(candidate[field_name]) + "]"

            # skip this field if it is a string field and its length is less than the chunk size
            if len(candidate[field_name]) < self.chunk_size:
                continue

            # chunk the content
            chunks = self.chunk_text(candidate[field_name], self.chunk_size)

            # compute embeddings for each chunk
            chunk_embeddings, chunk_embed_stats_lst = zip(*[self.compute_embedding(chunk) for chunk in chunks])

            # add cost of embedding each chunk to embed_stats
            for chunk_embed_stats in chunk_embed_stats_lst:
                embed_stats += chunk_embed_stats

            # select the top-k chunks
            sorted_chunks = sorted(
                zip(range(len(chunks)), chunks, chunk_embeddings),
                key=lambda tup: self.compute_similarity(query_embedding, tup[2]),
                reverse=True,
            )
            top_k_chunks = [(chunk_idx, chunk) for chunk_idx, chunk, _ in sorted_chunks[:self.num_chunks_per_field]]

            # sort the top-k chunks by their original index in the content, and join them with ellipses
            top_k_chunks = [chunk for _, chunk in sorted(top_k_chunks, key=lambda tup: tup[0])]
            candidate[field_name] = "...".join(top_k_chunks)

        return candidate, embed_stats

    def convert(self, candidate: DataRecord, fields: dict[str, Field]) -> tuple[dict[str, list], GenerationStats]:
        # set client
        self.client = OpenAI() if self.client is None else self.client

        # get the set of input fields to use for the convert operation
        input_fields = self.get_input_fields()
        output_fields = list(fields.keys())

        # lookup most relevant chunks for each field using embedding search
        candidate_copy = candidate.copy()
        candidate_copy, embed_stats = self.get_chunked_candidate(candidate_copy, input_fields, output_fields)

        # construct kwargs for generation
        gen_kwargs = {"project_cols": input_fields, "output_schema": self.output_schema}

        # generate outputs for all fields in a single query
        field_answers, _, generation_stats, _ = self.generator(candidate_copy, fields, **gen_kwargs)

        # NOTE: summing embedding stats with generation stats is messy because it will lead to misleading
        #       measurements of total_input_tokens and total_output_tokens. We should fix this in the future.
        #       The good news: as long as we compute the cost_per_record of each GenerationStats object correctly,
        #       then the total cost of the operation will be correct (which will roll-up to correctly computing
        #       the total cost of the operator, plan, and execution).
        #
        # combine stats from embedding with stats for generation
        generation_stats += embed_stats

        # if there was an error for any field, execute a conventional query on that field
        for field_name, answers in field_answers.items():
            if answers is None:
                single_field_answers, _, single_field_stats, _ = self.generator(candidate_copy, {field_name: fields[field_name]}, **gen_kwargs)
                field_answers.update(single_field_answers)
                generation_stats += single_field_stats

        return field_answers, generation_stats



"""This file contains utility phrases which are templated into many of our prompts."""

### FORMATTING INSTRUCTIONS ###
ONE_TO_ONE_OUTPUT_FORMAT_INSTRUCTION = "Remember, your answer must be a valid JSON dictionary. The dictionary should only have the specified output fields."
ONE_TO_MANY_OUTPUT_FORMAT_INSTRUCTION = "Remember, your answer must be a valid JSON list of dictionaries. The list may contain one or more dictionaries, and each dictionary should only have the specified output fields."

### REASONING INSTRUCTION FOR IMAGE PROMPTS ###
COT_REASONING_INSTRUCTION = """Let's think step-by-step in order to answer the question.

REASONING: """

COT_ANSWER_INSTRUCTION = """Let's think step-by-step in order to answer the question.

ANSWER: """



import os
from pathlib import Path

import pandas as pd

from palimpzest import constants
from palimpzest.core.data.datareaders import (
    DataReader,
    FileReader,
    HTMLFileDirectoryReader,
    ImageFileDirectoryReader,
    MemoryReader,
    PDFFileDirectoryReader,
    TextFileDirectoryReader,
    XLSFileDirectoryReader,
)


def get_local_source(path: str | Path, **kwargs) -> DataReader:
    """Return a DataReader for a local file or directory."""
    if os.path.isfile(path):
        return FileReader(path)

    elif os.path.isdir(path):
        if all([f.endswith(tuple(constants.IMAGE_EXTENSIONS)) for f in os.listdir(path)]):
            return ImageFileDirectoryReader(path)

        elif all([f.endswith(tuple(constants.PDF_EXTENSIONS)) for f in os.listdir(path)]):
            pdfprocessor = kwargs.get("pdfprocessor", constants.DEFAULT_PDF_PROCESSOR)
            file_cache_dir = kwargs.get("file_cache_dir", "/tmp")
            return PDFFileDirectoryReader(
                path=path, pdfprocessor=pdfprocessor, file_cache_dir=file_cache_dir
            )

        elif all([f.endswith(tuple(constants.XLS_EXTENSIONS)) for f in os.listdir(path)]):
            return XLSFileDirectoryReader(path)

        elif all([f.endswith(tuple(constants.HTML_EXTENSIONS)) for f in os.listdir(path)]):
            return HTMLFileDirectoryReader(path)

        else:
            return TextFileDirectoryReader(path)
    else:
        raise Exception(f"Path {path} is invalid. Does not point to a file or directory.")


def get_local_datareader(source: str | Path | list | pd.DataFrame, **kwargs) -> DataReader:
    """
    This helper function returns a `DataReader` object based on the `source` type.
    The returned `DataReader` object is guaranteed to have a schema.
    """
    if isinstance(source, (str, Path)):
        source = get_local_source(source, **kwargs)

    elif isinstance(source, (list, pd.DataFrame)):
        source = MemoryReader(source)

    else:
        raise Exception(f"Invalid source type: {type(source)}, We only support str, Path, list[dict], and pd.DataFrame")

    return source



import logging
import multiprocessing
from concurrent.futures import ThreadPoolExecutor, wait

from palimpzest.constants import PARALLEL_EXECUTION_SLEEP_INTERVAL_SECS
from palimpzest.core.data.dataclasses import PlanStats
from palimpzest.core.elements.records import DataRecord, DataRecordSet
from palimpzest.query.execution.execution_strategy import ExecutionStrategy
from palimpzest.query.operators.aggregate import AggregateOp
from palimpzest.query.operators.limit import LimitScanOp
from palimpzest.query.operators.physical import PhysicalOperator
from palimpzest.query.operators.scan import ScanPhysicalOp
from palimpzest.query.optimizer.plan import PhysicalPlan
from palimpzest.utils.progress import create_progress_manager

logger = logging.getLogger(__name__)

class ParallelExecutionStrategy(ExecutionStrategy):
    """
    A parallel execution strategy that processes data through a pipeline of operators using thread-based parallelism.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.max_workers = (
            self._get_parallel_max_workers()
            if self.max_workers is None
            else self.max_workers
        )

    def _get_parallel_max_workers(self):
        # for now, return the number of system CPUs;
        # in the future, we may want to consider the models the user has access to
        # and whether or not they will encounter rate-limits. If they will, we should
        # set the max workers in a manner that is designed to avoid hitting them.
        # Doing this "right" may require considering their logical, physical plan,
        # and tier status with LLM providers. It may also be worth dynamically
        # changing the max_workers in response to 429 errors.
        return max(int(0.8 * multiprocessing.cpu_count()), 1)

    def _any_queue_not_empty(self, queues: dict[str, list]) -> bool:
        """Helper function to check if any queue is not empty."""
        return any(len(queue) > 0 for queue in queues.values())

    def _upstream_ops_finished(self, plan: PhysicalPlan, op_idx: int, input_queues: dict[str, list], future_queues: dict[str, list]) -> bool:
        """Helper function to check if all upstream operators have finished processing their inputs."""
        for upstream_op_idx in range(op_idx):
            upstream_full_op_id = plan.operators[upstream_op_idx].get_full_op_id()
            if len(input_queues[upstream_full_op_id]) > 0 or len(future_queues[upstream_full_op_id]) > 0:
                return False

        return True

    def _process_future_results(self, operator: PhysicalOperator, future_queues: dict[str, list], plan_stats: PlanStats) -> list[DataRecord]:
        """
        Helper function which takes an operator, the future queues, and plan stats, and performs
        the updates to plan stats and progress manager before returning the results from the finished futures.
        """
        # get the op_id for the operator
        full_op_id = operator.get_full_op_id()

        # this function is called when the future queue is not empty
        # and the executor is not busy processing other futures
        done_futures, not_done_futures = wait(future_queues[full_op_id], timeout=PARALLEL_EXECUTION_SLEEP_INTERVAL_SECS)

        # add the unfinished futures back to the previous op's future queue
        future_queues[full_op_id] = list(not_done_futures)

        # add the finished futures to the input queue for this operator
        output_records = []
        for future in done_futures:
            record_set: DataRecordSet = future.result()
            records = record_set.data_records
            record_op_stats = record_set.record_op_stats
            num_outputs = sum(record.passed_operator for record in records)

            # update the progress manager
            self.progress_manager.incr(full_op_id, num_outputs=num_outputs, total_cost=record_set.get_total_cost())

            # update plan stats
            plan_stats.add_record_op_stats(record_op_stats)

            # add records to the cache
            self._add_records_to_cache(operator.target_cache_id, records)

            # add records which aren't filtered to the output records
            output_records.extend([record for record in records if record.passed_operator])
        
        return output_records

    def _execute_plan(
            self,
            plan: PhysicalPlan,
            input_queues: dict[str, list],
            future_queues: dict[str, list],
            plan_stats: PlanStats,
        ) -> tuple[list[DataRecord], PlanStats]:
        # process all of the input records using a thread pool
        output_records = []
        with ThreadPoolExecutor(max_workers=self.max_workers) as executor:
            logger.debug(f"Created thread pool with {self.max_workers} workers")

            # execute the plan until either:
            # 1. all records have been processed, or
            # 2. the final limit operation has completed (we break out of the loop if this happens)
            final_op = plan.operators[-1]
            while self._any_queue_not_empty(input_queues) or self._any_queue_not_empty(future_queues):
                for op_idx, operator in enumerate(plan.operators):
                    full_op_id = operator.get_full_op_id()

                    # get any finished futures from the previous operator and add them to the input queue for this operator
                    if not isinstance(operator, ScanPhysicalOp):
                        prev_operator = plan.operators[op_idx - 1]
                        records = self._process_future_results(prev_operator, future_queues, plan_stats)
                        input_queues[full_op_id].extend(records)

                    # for the final operator, add any finished futures to the output_records
                    if full_op_id == final_op.get_full_op_id():
                        records = self._process_future_results(operator, future_queues, plan_stats)
                        output_records.extend(records)

                    # if this operator does not have enough inputs to execute, then skip it
                    num_inputs = len(input_queues[full_op_id])
                    agg_op_not_ready = isinstance(operator, AggregateOp) and not self._upstream_ops_finished(plan, op_idx, input_queues, future_queues)
                    if num_inputs == 0 or agg_op_not_ready:
                        continue

                    # if this operator is an aggregate, process all the records in the input queue
                    if isinstance(operator, AggregateOp):
                        input_records = [input_queues[full_op_id].pop(0) for _ in range(num_inputs)]
                        future = executor.submit(operator, input_records)
                        future_queues[full_op_id].append(future)

                    else:
                        input_record = input_queues[full_op_id].pop(0)
                        future = executor.submit(operator, input_record)
                        future_queues[full_op_id].append(future)

                # break out of loop if the final operator is a LimitScanOp and we've reached its limit
                if isinstance(final_op, LimitScanOp) and len(output_records) == final_op.limit:
                    break

        # close the cache
        self._close_cache([op.target_cache_id for op in plan.operators])

        # finalize plan stats
        plan_stats.finish()

        return output_records, plan_stats


    def execute_plan(self, plan: PhysicalPlan):
        """Initialize the stats and execute the plan."""
        # for now, assert that the first operator in the plan is a ScanPhysicalOp
        assert isinstance(plan.operators[0], ScanPhysicalOp), "First operator in physical plan must be a ScanPhysicalOp"
        logger.info(f"Executing plan {plan.plan_id} with {self.max_workers} workers")
        logger.info(f"Plan Details: {plan}")

        # initialize plan stats
        plan_stats = PlanStats.from_plan(plan)
        plan_stats.start()

        # initialize input queues and future queues for each operation
        input_queues = self._create_input_queues(plan)
        future_queues = {op.get_full_op_id(): [] for op in plan.operators}

        # initialize and start the progress manager
        self.progress_manager = create_progress_manager(plan, num_samples=self.num_samples, progress=self.progress)
        self.progress_manager.start()

        # NOTE: we must handle progress manager outside of _exeecute_plan to ensure that it is shut down correctly;
        #       if we don't have the `finally:` branch, then program crashes can cause future program runs to fail
        #       because the progress manager cannot get a handle to the console 
        try:
            # execute plan
            output_records, plan_stats = self._execute_plan(plan, input_queues, future_queues, plan_stats)

        finally:
            # finish progress tracking
            self.progress_manager.finish()

        logger.info(f"Done executing plan: {plan.plan_id}")
        logger.debug(f"Plan stats: (plan_cost={plan_stats.total_plan_cost}, plan_time={plan_stats.total_plan_time})")

        return output_records, plan_stats


class SequentialParallelExecutionStrategy(ExecutionStrategy):
    """
    A parallel execution strategy that processes operators sequentially.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.max_workers = (
            self._get_parallel_max_workers()
            if self.max_workers is None
            else self.max_workers
        )

    def _get_parallel_max_workers(self):
        # for now, return the number of system CPUs;
        # in the future, we may want to consider the models the user has access to
        # and whether or not they will encounter rate-limits. If they will, we should
        # set the max workers in a manner that is designed to avoid hitting them.
        # Doing this "right" may require considering their logical, physical plan,
        # and tier status with LLM providers. It may also be worth dynamically
        # changing the max_workers in response to 429 errors.
        return max(int(0.8 * multiprocessing.cpu_count()), 1)

    def _any_queue_not_empty(self, queues: dict[str, list]) -> bool:
        """Helper function to check if any queue is not empty."""
        return any(len(queue) > 0 for queue in queues.values())

    def _upstream_ops_finished(self, plan: PhysicalPlan, op_idx: int, input_queues: dict[str, list], future_queues: dict[str, list]) -> bool:
        """Helper function to check if all upstream operators have finished processing their inputs."""
        for upstream_op_idx in range(op_idx):
            upstream_full_op_id = plan.operators[upstream_op_idx].get_full_op_id()
            if len(input_queues[upstream_full_op_id]) > 0 or len(future_queues[upstream_full_op_id]) > 0:
                return False

        return True

    def _process_future_results(self, operator: PhysicalOperator, future_queues: dict[str, list], plan_stats: PlanStats) -> list[DataRecord]:
        """
        Helper function which takes an operator, the future queues, and plan stats, and performs
        the updates to plan stats and progress manager before returning the results from the finished futures.
        """
        # get the op_id for the operator
        full_op_id = operator.get_full_op_id()

        # this function is called when the future queue is not empty
        # and the executor is not busy processing other futures
        done_futures, not_done_futures = wait(future_queues[full_op_id], timeout=PARALLEL_EXECUTION_SLEEP_INTERVAL_SECS)

        # add the unfinished futures back to the previous op's future queue
        future_queues[full_op_id] = list(not_done_futures)

        # add the finished futures to the input queue for this operator
        output_records = []
        for future in done_futures:
            record_set: DataRecordSet = future.result()
            records = record_set.data_records
            record_op_stats = record_set.record_op_stats
            num_outputs = sum(record.passed_operator for record in records)

            # update the progress manager
            self.progress_manager.incr(full_op_id, num_outputs=num_outputs, total_cost=record_set.get_total_cost())

            # update plan stats
            plan_stats.add_record_op_stats(record_op_stats)

            # add records to the cache
            self._add_records_to_cache(operator.target_cache_id, records)

            # add records which aren't filtered to the output records
            output_records.extend([record for record in records if record.passed_operator])
        
        return output_records

    def _execute_plan(
            self,
            plan: PhysicalPlan,
            input_queues: dict[str, list],
            future_queues: dict[str, list],
            plan_stats: PlanStats,
        ) -> tuple[list[DataRecord], PlanStats]:
        # process all of the input records using a thread pool
        output_records = []
        with ThreadPoolExecutor(max_workers=self.max_workers) as executor:
            logger.debug(f"Created thread pool with {self.max_workers} workers")

            # execute the plan until either:
            # 1. all records have been processed, or
            # 2. the final limit operation has completed (we break out of the loop if this happens)
            final_op = plan.operators[-1]
            for op_idx, operator in enumerate(plan.operators):
                full_op_id = operator.get_full_op_id()
                input_queue = input_queues[full_op_id]

                # if this operator is an aggregate, process all the records in the input queue
                if isinstance(operator, AggregateOp):
                    num_inputs = len(input_queue)
                    input_records = [input_queue.pop(0) for _ in range(num_inputs)]
                    future = executor.submit(operator, input_records)
                    future_queues[full_op_id].append(future)

                else:
                    while len(input_queue) > 0:
                        input_record = input_queue.pop(0)
                        future = executor.submit(operator, input_record)
                        future_queues[full_op_id].append(future)

                # block until all futures for this operator have completed; and add finished futures to next operator's input
                while len(future_queues[full_op_id]) > 0:
                    records = self._process_future_results(operator, future_queues, plan_stats)

                    # get any finished futures from the previous operator and add them to the input queue for this operator
                    if full_op_id != final_op.get_full_op_id():
                        next_op_id = plan.operators[op_idx + 1].get_full_op_id()
                        input_queues[next_op_id].extend(records)

                    # for the final operator, add any finished futures to the output_records
                    else:
                        output_records.extend(records)

                        # break out of loop if the final operator is a LimitScanOp and we've reached its limit
                        if isinstance(final_op, LimitScanOp) and len(output_records) == final_op.limit:
                            break

        # close the cache
        self._close_cache([op.target_cache_id for op in plan.operators])

        # finalize plan stats
        plan_stats.finish()

        return output_records, plan_stats


    def execute_plan(self, plan: PhysicalPlan):
        """Initialize the stats and execute the plan."""
        # for now, assert that the first operator in the plan is a ScanPhysicalOp
        assert isinstance(plan.operators[0], ScanPhysicalOp), "First operator in physical plan must be a ScanPhysicalOp"
        logger.info(f"Executing plan {plan.plan_id} with {self.max_workers} workers")
        logger.info(f"Plan Details: {plan}")

        # initialize plan stats
        plan_stats = PlanStats.from_plan(plan)
        plan_stats.start()

        # initialize input queues and future queues for each operation
        input_queues = self._create_input_queues(plan)
        future_queues = {op.get_full_op_id(): [] for op in plan.operators}

        # initialize and start the progress manager
        self.progress_manager = create_progress_manager(plan, num_samples=self.num_samples, progress=self.progress)
        self.progress_manager.start()

        # NOTE: we must handle progress manager outside of _exeecute_plan to ensure that it is shut down correctly;
        #       if we don't have the `finally:` branch, then program crashes can cause future program runs to fail
        #       because the progress manager cannot get a handle to the console 
        try:
            # execute plan
            output_records, plan_stats = self._execute_plan(plan, input_queues, future_queues, plan_stats)

        finally:
            # finish progress tracking
            self.progress_manager.finish()

        logger.info(f"Done executing plan: {plan.plan_id}")
        logger.debug(f"Plan stats: (plan_cost={plan_stats.total_plan_cost}, plan_time={plan_stats.total_plan_time})")

        return output_records, plan_stats



import logging

import numpy as np

from palimpzest.core.data.dataclasses import OperatorStats, SentinelPlanStats
from palimpzest.core.elements.records import DataRecord, DataRecordSet
from palimpzest.policy import Policy
from palimpzest.query.execution.execution_strategy import SentinelExecutionStrategy
from palimpzest.query.operators.filter import FilterOp
from palimpzest.query.operators.physical import PhysicalOperator
from palimpzest.query.operators.scan import ScanPhysicalOp
from palimpzest.query.optimizer.plan import SentinelPlan
from palimpzest.utils.progress import create_progress_manager

logger = logging.getLogger(__name__)

class OpFrontier:
    """
    This class represents the set of operators which are currently in the frontier for a given logical operator.
    Each operator in the frontier is an instance of a PhysicalOperator which either:

    1. lies on the Pareto frontier of the set of sampled operators, or
    2. has been sampled fewer than j times
    """

    def __init__(self, op_set: list[PhysicalOperator], source_indices: list[int], k: int, j: int, seed: int, policy: Policy, priors: dict | None = None):
        # set k and j, which are the initial number of operators in the frontier and the
        # initial number of records to sample for each frontier operator
        self.k = min(k, len(op_set))
        self.j = min(j, len(source_indices))

        # store the policy that we are optimizing under
        self.policy = policy

        # store the prior beliefs on operator performance (if provided)
        self.priors = priors

        # get order in which we will sample physical operators for this logical operator
        sample_op_indices = self._get_op_index_order(op_set, seed)

        # construct the initial set of frontier and reservoir operators
        self.frontier_ops = [op_set[sample_idx] for sample_idx in sample_op_indices[:self.k]]
        self.reservoir_ops = [op_set[sample_idx] for sample_idx in sample_op_indices[self.k:]]
        self.off_frontier_ops: list[PhysicalOperator] = []

        # store the order in which we will sample the source records
        self.source_indices = source_indices

        # keep track of the source ids processed by each physical operator
        self.full_op_id_to_sources_processed = {op.get_full_op_id(): set() for op in op_set}

        # set the initial inputs for this logical operator
        is_scan_op = isinstance(op_set[0], ScanPhysicalOp)
        self.source_idx_to_input = {source_idx: [source_idx] for source_idx in self.source_indices} if is_scan_op else {}

        # boolean indication of whether this is a logical filter
        self.is_filter_op = isinstance(op_set[0], FilterOp)

    def get_frontier_ops(self) -> list[PhysicalOperator]:
        """
        Returns the set of frontier operators for this OpFrontier.
        """
        return self.frontier_ops

    def _compute_op_id_to_pareto_distance(self, priors: dict[str, dict[str, float]]) -> dict[str, float]:
        """
        Return l2-distance for each operator from the pareto frontier.
        """
        # get the dictionary representation of this poicy
        policy_dict = self.policy.get_dict()

        # compute the pareto optimal set of operators
        pareto_op_set = set()
        for op_id, metrics in priors.items():
            cost, time, quality = metrics["cost"], metrics["time"], metrics["quality"]
            pareto_frontier = True

            # check if any other operator dominates op_id
            for other_op_id, other_metrics in priors.items():
                other_cost, other_time, other_quality = other_metrics["cost"], other_metrics["time"], other_metrics["quality"]
                if op_id == other_op_id:
                    continue

                # if op_id is dominated by other_op_id, set pareto_frontier = False and break
                # NOTE: here we use a strict inequality (instead of the usual <= or >=) because
                #       all ops which have equal cost / time / quality / sel. should not be
                #       filtered out from sampling by our logic in this function
                cost_dominated = True if policy_dict["cost"] == 0.0 else other_cost < cost
                time_dominated = True if policy_dict["time"] == 0.0 else other_time < time
                quality_dominated = True if policy_dict["quality"] == 0.0 else other_quality > quality
                if cost_dominated and time_dominated and quality_dominated:
                    pareto_frontier = False
                    break

            # add op_id to pareto frontier if it's not dominated
            if pareto_frontier:
                pareto_op_set.add(op_id)

        # compute the shortest distance from each operator to the pareto frontier
        op_id_to_pareto_distance = {}
        for op_id, metrics in priors.items():
            # set distance to 0.0 if this operator is on the pareto frontier
            if op_id in pareto_op_set:
                op_id_to_pareto_distance[op_id] = 0.0
                continue

            # otherwise, compute min_dist to pareto operators
            min_dist = None
            cost, time, quality = metrics["cost"], metrics["time"], metrics["quality"]
            for pareto_op_id in pareto_op_set:
                pareto_cost, pareto_time, pareto_quality = priors[pareto_op_id]["cost"], priors[pareto_op_id]["time"], priors[pareto_op_id]["quality"]

                cost_dist_squared = 0.0 if policy_dict["cost"] == 0.0 else (cost - pareto_cost) ** 2
                time_dist_squared = 0.0 if policy_dict["time"] == 0.0 else (time - pareto_time) ** 2
                quality_dist_squared = 0.0 if policy_dict["quality"] == 0.0 else (quality - pareto_quality) ** 2
                dist = np.sqrt(cost_dist_squared + time_dist_squared + quality_dist_squared)
                if min_dist is None or dist < min_dist:
                    min_dist = dist

            # set minimum distance for this operator
            op_id_to_pareto_distance[op_id] = min_dist
        
        return op_id_to_pareto_distance

    def _get_op_index_order(self, op_set: list[PhysicalOperator], seed: int) -> list[int]:
        """
        Returns a list of indices for the operators in the op_set.
        """
        if self.priors is None or any([op_id not in self.priors for op_id in map(lambda op: op.get_op_id(), op_set)]):
            rng = np.random.default_rng(seed=seed)
            op_indices = np.arange(len(op_set))
            rng.shuffle(op_indices)
            return op_indices

        # NOTE: self.priors is a dictionary with format:
        # {op_id: {"quality": quality, "cost": cost, "time": time}}

        # compute mean and std. dev. for each field
        qualities = [op_priors["quality"] for op_priors in self.priors.values()]
        costs = [op_priors["cost"] for op_priors in self.priors.values()]
        times = [op_priors["time"] for op_priors in self.priors.values()]
        metric_to_mean = {"quality": np.mean(qualities), "cost": np.mean(costs), "time": np.mean(times)}
        metric_to_std = {"quality": np.std(qualities), "cost": np.std(costs), "time": np.std(times)}

        # normalize the scale of each field to be the same
        for _, op_priors in self.priors.items():
            for metric, value in op_priors.items():
                if metric_to_std[metric] == 0.0:
                    op_priors[metric] = metric_to_mean[metric]
                else:
                    op_priors[metric] = (value - metric_to_mean[metric]) / metric_to_std[metric]

        # then, we compute the l2-distance from the pareto frontier for each operator
        op_id_to_distance = self._compute_op_id_to_pareto_distance(self.priors)

        # compute tuple for every operator, invert quality so ascending sort puts
        # best operator first: (op_id, dist, -1 * quality, cost, time);
        op_tuples = []
        for op in op_set:
            op_id = op.get_op_id()
            op_priors = self.priors[op_id]
            op_tuple = (op_id, op_id_to_distance[op_id], -1 * op_priors["quality"], op_priors["cost"], op_priors["time"])
            op_tuples.append(op_tuple)

        # sort tuples on distance, then second dim
        second_dim_idx = None
        if self.policy.get_primary_metric() == "quality":
            second_dim_idx = 2
        elif self.policy.get_primary_metric() == "cost":
            second_dim_idx = 3
        elif self.policy.get_primary_metric() == "time":
            second_dim_idx = 4

        # sort based on distance from pareto frontier; break ties with performance on max / min metric
        op_tuples = sorted(op_tuples, key=lambda x: (x[1], x[second_dim_idx]))

        # return final list of op indices in sample order
        op_id_to_idx = {op.get_op_id(): idx for idx, op in enumerate(op_set)}
        op_indices = [op_id_to_idx[op_tuple[0]] for op_tuple in op_tuples]

        return op_indices

    def _get_op_source_idx_pairs(self) -> list[tuple[PhysicalOperator, int]]:
        """
        Returns a list of tuples for (op, source_idx) which this operator needs to execute
        in the next iteration.
        """
        op_source_idx_pairs = []
        for op in self.frontier_ops:
            # execute new operators on first j source indices, and previously sampled operators on one additional source_idx
            num_processed = len(self.full_op_id_to_sources_processed[op.get_full_op_id()])
            num_new_samples = 1 if num_processed > 0 else self.j
            num_new_samples = min(num_new_samples, len(self.source_indices) - num_processed)
            assert num_new_samples >= 0, "Number of new samples must be non-negative"

            # construct list of inputs by looking up the input for the given source_idx
            samples_added = 0
            for source_idx in self.source_indices:
                if source_idx in self.full_op_id_to_sources_processed[op.get_full_op_id()]:
                    continue

                if samples_added == num_new_samples:
                    break

                # construct the (op, source_idx) for this source_idx
                op_source_idx_pairs.append((op, source_idx))
                samples_added += 1

        return op_source_idx_pairs

    def get_source_indices_for_next_iteration(self) -> set[int]:
        """
        Returns the set of source indices which need to be sampled for the next iteration.
        """
        op_source_idx_pairs = self._get_op_source_idx_pairs()
        return set(map(lambda tup: tup[1], op_source_idx_pairs))

    def get_frontier_op_input_pairs(self, source_indices_to_sample: set[int], max_quality_op: PhysicalOperator) -> list[PhysicalOperator, DataRecord | int | None]:
        """
        Returns the list of frontier operators and their next input to process. If there are
        any indices in `source_indices_to_sample` which this operator does not sample on its own, then
        we also have this frontier process that source_idx's input with its max quality operator.
        """
        # get the list of (op, source_idx) pairs which this operator needs to execute
        op_source_idx_pairs = self._get_op_source_idx_pairs()

        # if there are any source_idxs in source_indices_to_sample which are not sampled
        # by this operator, apply the max quality operator (and any other frontier operators
        # with no samples)
        sampled_source_indices = set(map(lambda tup: tup[1], op_source_idx_pairs))
        unsampled_source_indices = source_indices_to_sample - sampled_source_indices
        for source_idx in unsampled_source_indices:
            op_source_idx_pairs.append((max_quality_op, source_idx))
            for op in self.frontier_ops:
                if len(self.full_op_id_to_sources_processed[op.get_full_op_id()]) == 0 and op.get_full_op_id() != max_quality_op.get_full_op_id():
                    op_source_idx_pairs.append((op, source_idx))

        # fetch the corresponding (op, input) pairs
        op_input_pairs = [
            (op, input)
            for op, source_idx in op_source_idx_pairs
            for input in self.source_idx_to_input[source_idx]
        ]

        return op_input_pairs

    def update_frontier(self, logical_op_id: str, plan_stats: SentinelPlanStats) -> None:
        """
        Update the set of frontier operators, pulling in new ones from the reservoir as needed.
        This function will:
        1. Compute the mean, LCB, and UCB for the cost, time, quality, and selectivity of each frontier operator
        2. Compute the pareto optimal set of frontier operators (using the mean values)
        3. Update the frontier and reservoir sets of operators based on their LCB/UCB overlap with the pareto frontier
        """
        # NOTE: downstream operators may end up re-computing the same record_id with a diff. input as upstream
        #       upstream operators change; in this case, we de-duplicate record_op_stats with identical record_ids
        #       and keep the one with the maximum quality
        # get a mapping from full_op_id --> list[RecordOpStats]
        full_op_id_to_op_stats: dict[str, OperatorStats] = plan_stats.operator_stats.get(logical_op_id, {})
        full_op_id_to_record_op_stats = {}
        for full_op_id, op_stats in full_op_id_to_op_stats.items():
            # skip over operators which have not been sampled
            if len(op_stats.record_op_stats_lst) == 0:
                continue

            # compute mapping from record_id to highest quality record op stats
            record_id_to_max_quality_record_op_stats = {}
            for record_op_stats in op_stats.record_op_stats_lst:
                record_id = record_op_stats.record_id
                if record_id not in record_id_to_max_quality_record_op_stats:  # noqa: SIM114
                    record_id_to_max_quality_record_op_stats[record_id] = record_op_stats

                elif record_op_stats.quality > record_id_to_max_quality_record_op_stats[record_id].quality:
                    record_id_to_max_quality_record_op_stats[record_id] = record_op_stats

            # compute final list of record op stats
            full_op_id_to_record_op_stats[full_op_id] = list(record_id_to_max_quality_record_op_stats.values())

        # compute mapping of physical op to num samples and total samples drawn;
        # also update the set of source indices which have been processed by each physical operator
        full_op_id_to_num_samples, total_num_samples = {}, 0
        for full_op_id, record_op_stats_lst in full_op_id_to_record_op_stats.items():
            # update the set of source indices processed
            for record_op_stats in record_op_stats_lst:
                self.full_op_id_to_sources_processed[full_op_id].add(record_op_stats.record_source_idx)

            # compute the number of samples as the number of source indices processed
            num_samples = len(self.full_op_id_to_sources_processed[full_op_id])
            full_op_id_to_num_samples[full_op_id] = num_samples
            total_num_samples += num_samples

        # compute avg. selectivity, cost, time, and quality for each physical operator
        def total_output(record_op_stats_lst):
            return sum([record_op_stats.passed_operator for record_op_stats in record_op_stats_lst])

        def total_input(record_op_stats_lst):
            return len(set([record_op_stats.record_parent_id for record_op_stats in record_op_stats_lst]))

        full_op_id_to_mean_selectivity = {
            full_op_id: total_output(record_op_stats_lst) / total_input(record_op_stats_lst)
            for full_op_id, record_op_stats_lst in full_op_id_to_record_op_stats.items()
        }
        full_op_id_to_mean_cost = {
            full_op_id: np.mean([record_op_stats.cost_per_record for record_op_stats in record_op_stats_lst])
            for full_op_id, record_op_stats_lst in full_op_id_to_record_op_stats.items()
        }
        full_op_id_to_mean_time = {
            full_op_id: np.mean([record_op_stats.time_per_record for record_op_stats in record_op_stats_lst])
            for full_op_id, record_op_stats_lst in full_op_id_to_record_op_stats.items()
        }
        full_op_id_to_mean_quality = {
            full_op_id: np.mean([record_op_stats.quality for record_op_stats in record_op_stats_lst])
            for full_op_id, record_op_stats_lst in full_op_id_to_record_op_stats.items()
        }

        # # compute average, LCB, and UCB of each operator; the confidence bounds depend upon
        # # the computation of the alpha parameter, which we scale to be 0.5 * the mean (of means)
        # # of the metric across all operators in this operator set
        # cost_alpha = 0.5 * np.mean([mean_cost for mean_cost in full_op_id_to_mean_cost.values()])
        # time_alpha = 0.5 * np.mean([mean_time for mean_time in full_op_id_to_mean_time.values()])
        # quality_alpha = 0.5 * np.mean([mean_quality for mean_quality in full_op_id_to_mean_quality.values()])
        # selectivity_alpha = 0.5 * np.mean([mean_selectivity for mean_selectivity in full_op_id_to_mean_selectivity.values()])
        cost_alpha = 0.5 * (np.max(list(full_op_id_to_mean_cost.values())) - np.min(list(full_op_id_to_mean_cost.values())))
        time_alpha = 0.5 * (np.max(list(full_op_id_to_mean_time.values())) - np.min(list(full_op_id_to_mean_time.values())))
        quality_alpha = 0.5 * (np.max(list(full_op_id_to_mean_quality.values())) - np.min(list(full_op_id_to_mean_quality.values())))
        selectivity_alpha = 0.5 * (np.max(list(full_op_id_to_mean_selectivity.values())) - np.min(list(full_op_id_to_mean_selectivity.values())))

        # compute metrics for each physical operator
        op_metrics = {}
        for full_op_id in full_op_id_to_record_op_stats:
            sample_ratio = np.sqrt(np.log(total_num_samples) / full_op_id_to_num_samples[full_op_id])
            exploration_terms = np.array([cost_alpha * sample_ratio, time_alpha * sample_ratio, quality_alpha * sample_ratio, selectivity_alpha * sample_ratio])
            mean_terms = (full_op_id_to_mean_cost[full_op_id], full_op_id_to_mean_time[full_op_id], full_op_id_to_mean_quality[full_op_id], full_op_id_to_mean_selectivity[full_op_id])

            # NOTE: we could clip these; however I will not do so for now to allow for arbitrary quality metric(s)
            lcb_terms = mean_terms - exploration_terms
            ucb_terms = mean_terms + exploration_terms
            op_metrics[full_op_id] = {"mean": mean_terms, "lcb": lcb_terms, "ucb": ucb_terms}

        # get the tuple representation of this policy
        policy_dict = self.policy.get_dict()

        # compute the pareto optimal set of operators
        pareto_op_set = set()
        for full_op_id, metrics in op_metrics.items():
            cost, time, quality, selectivity = metrics["mean"]
            pareto_frontier = True

            # check if any other operator dominates full_op_id
            for other_full_op_id, other_metrics in op_metrics.items():
                other_cost, other_time, other_quality, other_selectivity = other_metrics["mean"]
                if full_op_id == other_full_op_id:
                    continue

                # if full_op_id is dominated by other_full_op_id, set pareto_frontier = False and break
                # NOTE: here we use a strict inequality (instead of the usual <= or >=) because
                #       all ops which have equal cost / time / quality / sel. should not be
                #       filtered out from sampling by our logic in this function
                cost_dominated = True if policy_dict["cost"] == 0.0 else other_cost < cost
                time_dominated = True if policy_dict["time"] == 0.0 else other_time < time
                quality_dominated = True if policy_dict["quality"] == 0.0 else other_quality > quality
                selectivity_dominated = True if not self.is_filter_op else other_selectivity < selectivity
                if cost_dominated and time_dominated and quality_dominated and selectivity_dominated:
                    pareto_frontier = False
                    break

            # add full_op_id to pareto frontier if it's not dominated
            if pareto_frontier:
                pareto_op_set.add(full_op_id)

        # iterate over op metrics and compute the new frontier set of operators
        new_frontier_full_op_ids = set()
        for full_op_id, metrics in op_metrics.items():

            # if this op is fully sampled, do not keep it on the frontier
            if full_op_id_to_num_samples[full_op_id] == len(self.source_indices):
                continue

            # if this op is pareto optimal keep it in our frontier ops
            if full_op_id in pareto_op_set:
                new_frontier_full_op_ids.add(full_op_id)
                continue

            # otherwise, if this op overlaps with an op on the pareto frontier, keep it in our frontier ops
            # NOTE: for now, we perform an optimistic comparison with the ucb/lcb
            pareto_frontier = True
            op_cost, op_time, _, op_selectivity = metrics["lcb"]
            op_quality = metrics["ucb"][2]
            for pareto_full_op_id in pareto_op_set:
                pareto_cost, pareto_time, _, pareto_selectivity = op_metrics[pareto_full_op_id]["ucb"]
                pareto_quality = op_metrics[pareto_full_op_id]["lcb"][2]

                # if full_op_id is dominated by pareto_full_op_id, set pareto_frontier = False and break
                cost_dominated = True if policy_dict["cost"] == 0.0 else pareto_cost <= op_cost
                time_dominated = True if policy_dict["time"] == 0.0 else pareto_time <= op_time
                quality_dominated = True if policy_dict["quality"] == 0.0 else pareto_quality >= op_quality
                selectivity_dominated = True if not self.is_filter_op else pareto_selectivity <= op_selectivity
                if cost_dominated and time_dominated and quality_dominated and selectivity_dominated:
                    pareto_frontier = False
                    break

            # add full_op_id to pareto frontier if it's not dominated
            if pareto_frontier:
                new_frontier_full_op_ids.add(full_op_id)

        # for operators that were in the frontier, keep them in the frontier if they
        # are still pareto optimal, otherwise, move them to the end of the reservoir
        new_frontier_ops = []
        for op in self.frontier_ops:
            if op.get_full_op_id() in new_frontier_full_op_ids:
                new_frontier_ops.append(op)
            else:
                self.off_frontier_ops.append(op)

        # if there are operators we previously sampled which are now back on the frontier
        # add them to the frontier, otherwise, put them back in the off_frontier_ops
        new_off_frontier_ops = []
        for op in self.off_frontier_ops:
            if op.get_full_op_id() in new_frontier_full_op_ids:
                new_frontier_ops.append(op)
            else:
                new_off_frontier_ops.append(op)

        # finally, if we have fewer than k operators in the frontier, sample new operators
        # from the reservoir and put them in the frontier
        while len(new_frontier_ops) < self.k and len(self.reservoir_ops) > 0:
            new_op = self.reservoir_ops.pop(0)
            new_frontier_ops.append(new_op)

        # update the frontier and off frontier ops
        self.frontier_ops = new_frontier_ops
        self.off_frontier_ops = new_off_frontier_ops

    def pick_highest_quality_output(self, record_sets: list[DataRecordSet]) -> DataRecordSet:
        # if there's only one operator in the set, we return its record_set
        if len(record_sets) == 1:
            return record_sets[0]

        # NOTE: I don't like that this assumes the models are consistent in
        #       how they order their record outputs for one-to-many converts;
        #       eventually we can try out more robust schemes to account for
        #       differences in ordering
        # aggregate records at each index in the response
        idx_to_records = {}
        for record_set in record_sets:
            for idx in range(len(record_set)):
                record, record_op_stats = record_set[idx], record_set.record_op_stats[idx]
                if idx not in idx_to_records:
                    idx_to_records[idx] = [(record, record_op_stats)]
                else:
                    idx_to_records[idx].append((record, record_op_stats))

        # compute highest quality answer at each index
        out_records = []
        out_record_op_stats = []
        for idx in range(len(idx_to_records)):
            records_lst, record_op_stats_lst = zip(*idx_to_records[idx])
            max_quality_record, max_quality = records_lst[0], record_op_stats_lst[0].quality
            max_quality_stats = record_op_stats_lst[0]
            for record, record_op_stats in zip(records_lst[1:], record_op_stats_lst[1:]):
                record_quality = record_op_stats.quality
                if record_quality > max_quality:
                    max_quality_record = record
                    max_quality = record_quality
                    max_quality_stats = record_op_stats
            out_records.append(max_quality_record)
            out_record_op_stats.append(max_quality_stats)

        # create and return final DataRecordSet
        return DataRecordSet(out_records, out_record_op_stats)

    def update_inputs(self, source_idx_to_record_sets: dict[int, DataRecordSet]):
        """
        Update the inputs for this logical operator based on the outputs of the previous logical operator.
        """
        for source_idx, record_sets in source_idx_to_record_sets.items():
            input = []
            max_quality_record_set = self.pick_highest_quality_output(record_sets)
            for record in max_quality_record_set:
                input.append(record if record.passed_operator else None)

            self.source_idx_to_input[source_idx] = input


# TODO: post-submission we will need to modify this to:
# - submit all inputs for aggregate operators
# - handle limits
class MABExecutionStrategy(SentinelExecutionStrategy):
    """
    This class implements the Multi-Armed Bandit (MAB) execution strategy for SentinelQueryProcessors.

    NOTE: the number of samples will slightly exceed the sample_budget if the number of operator
    calls does not perfectly match the sample_budget. This may cause some minor discrepancies with
    the progress manager as a result.
    """
    def _get_max_quality_op(self, logical_op_id: str, op_frontiers: dict[str, OpFrontier], plan_stats: SentinelPlanStats) -> PhysicalOperator:
        """
        Returns the operator in the frontier with the highest (estimated) quality.
        """
        # get the operators in the frontier set for this logical_op_id
        frontier_ops = op_frontiers[logical_op_id].get_frontier_ops()

        # get a mapping from full_op_id --> list[RecordOpStats]
        full_op_id_to_op_stats: dict[str, OperatorStats] = plan_stats.operator_stats.get(logical_op_id, {})
        full_op_id_to_record_op_stats = {
            full_op_id: op_stats.record_op_stats_lst
            for full_op_id, op_stats in full_op_id_to_op_stats.items()
        }

        # iterate over the frontier ops and return the one with the highest quality
        max_quality_op, max_avg_quality = None, None
        for op in frontier_ops:
            op_quality_stats = []
            if op.get_full_op_id() in full_op_id_to_record_op_stats:
                op_quality_stats = [record_op_stats.quality for record_op_stats in full_op_id_to_record_op_stats[op.get_full_op_id()]]
            avg_op_quality = sum(op_quality_stats) / len(op_quality_stats) if len(op_quality_stats) > 0 else 0.0
            if max_avg_quality is None or avg_op_quality > max_avg_quality:
                max_quality_op = op
                max_avg_quality = avg_op_quality

        return max_quality_op

    def _execute_sentinel_plan(
            self,
            plan: SentinelPlan,
            op_frontiers: dict[str, OpFrontier],
            expected_outputs: dict[int, dict] | None,
            plan_stats: SentinelPlanStats,
        ) -> SentinelPlanStats:
        # sample records and operators and update the frontiers
        samples_drawn = 0
        while samples_drawn < self.sample_budget:
            # pre-compute the set of source indices which will need to be sampled
            source_indices_to_sample = set()
            for op_frontier in op_frontiers.values():
                source_indices = op_frontier.get_source_indices_for_next_iteration()
                source_indices_to_sample.update(source_indices)

            # execute operator sets in sequence
            for op_idx, (logical_op_id, op_set) in enumerate(plan):
                # use the execution cache to determine the maximum quality operator for this logical_op_id
                max_quality_op = self._get_max_quality_op(logical_op_id, op_frontiers, plan_stats)

                # TODO: can have None as an operator if _get_max_quality_op returns None
                # get frontier ops and their next input
                frontier_op_input_pairs = op_frontiers[logical_op_id].get_frontier_op_input_pairs(source_indices_to_sample, max_quality_op)
                frontier_op_input_pairs = list(filter(lambda tup: tup[1] is not None, frontier_op_input_pairs))

                # break out of the loop if frontier_op_input_pairs is empty, as this means all records have been filtered out
                if len(frontier_op_input_pairs) == 0:
                    break

                # run sampled operators on sampled inputs and update the number of samples drawn
                source_idx_to_record_set_tuples, num_llm_ops = self._execute_op_set(frontier_op_input_pairs)
                samples_drawn += num_llm_ops

                # FUTURE TODO: have this return the highest quality record set simply based on our posterior (or prior) belief on operator quality
                # get the target record set for each source_idx
                source_idx_to_target_record_set = self._get_target_record_sets(logical_op_id, source_idx_to_record_set_tuples, expected_outputs)

                # FUTURE TODO: move this outside of the loop (i.e. assume we only get quality label(s) after executing full program)
                # score the quality of each generated output
                physical_op_cls = op_set[0].__class__
                source_idx_to_all_record_sets = {
                    source_idx: [record_set for record_set, _, _ in record_set_tuples]
                    for source_idx, record_set_tuples in source_idx_to_record_set_tuples.items()
                }
                source_idx_to_all_record_sets = self._score_quality(physical_op_cls, source_idx_to_all_record_sets, source_idx_to_target_record_set)

                # flatten the lists of newly computed records and record_op_stats
                source_idx_to_new_record_sets = {
                    source_idx: [record_set for record_set, _, is_new in record_set_tuples if is_new]
                    for source_idx, record_set_tuples in source_idx_to_record_set_tuples.items()
                }
                new_records, new_record_op_stats = self._flatten_record_sets(source_idx_to_new_record_sets)

                # update the number of samples drawn for each operator

                # update plan stats
                plan_stats.add_record_op_stats(new_record_op_stats)

                # add records (which are not filtered) to the cache, if allowed
                self._add_records_to_cache(logical_op_id, new_records)

                # FUTURE TODO: simply set input based on source_idx_to_target_record_set (b/c we won't have scores computed)
                # provide the champion record sets as inputs to the next logical operator
                if op_idx + 1 < len(plan):
                    next_logical_op_id = plan.logical_op_ids[op_idx + 1]
                    op_frontiers[next_logical_op_id].update_inputs(source_idx_to_all_record_sets)

                # update the (pareto) frontier for each set of operators
                op_frontiers[logical_op_id].update_frontier(logical_op_id, plan_stats)

            # FUTURE TODO: score op quality based on final outputs

        # close the cache
        self._close_cache(plan.logical_op_ids)

        # finalize plan stats
        plan_stats.finish()

        return plan_stats


    def execute_sentinel_plan(self, plan: SentinelPlan, expected_outputs: dict[int, dict] | None):
        # for now, assert that the first operator in the plan is a ScanPhysicalOp
        assert all(isinstance(op, ScanPhysicalOp) for op in plan.operator_sets[0]), "First operator in physical plan must be a ScanPhysicalOp"
        logger.info(f"Executing plan {plan.plan_id} with {self.max_workers} workers")
        logger.info(f"Plan Details: {plan}")

        # initialize plan stats
        plan_stats = SentinelPlanStats.from_plan(plan)
        plan_stats.start()

        # shuffle the indices of records to sample
        total_num_samples = len(self.val_datasource)
        shuffled_source_indices = [int(idx) for idx in np.arange(total_num_samples)]
        self.rng.shuffle(shuffled_source_indices)

        # initialize frontier for each logical operator
        op_frontiers = {
            logical_op_id: OpFrontier(op_set, shuffled_source_indices, self.k, self.j, self.seed, self.policy, self.priors)
            for logical_op_id, op_set in plan
        }

        # initialize and start the progress manager
        self.progress_manager = create_progress_manager(plan, sample_budget=self.sample_budget, progress=self.progress)
        self.progress_manager.start()

        # NOTE: we must handle progress manager outside of _exeecute_sentinel_plan to ensure that it is shut down correctly;
        #       if we don't have the `finally:` branch, then program crashes can cause future program runs to fail because
        #       the progress manager cannot get a handle to the console 
        try:
            # execute sentinel plan by sampling records and operators
            plan_stats = self._execute_sentinel_plan(plan, op_frontiers, expected_outputs, plan_stats)

        finally:
            # finish progress tracking
            self.progress_manager.finish()

        logger.info(f"Done executing sentinel plan: {plan.plan_id}")
        logger.debug(f"Plan stats: (plan_cost={plan_stats.total_plan_cost}, plan_time={plan_stats.total_plan_time})")

        return plan_stats



### This file contains constants used by Palimpzest ###
import os
from enum import Enum


# ENUMS
class Model(str, Enum):
    """
    Model describes the underlying LLM which should be used to perform some operation
    which requires invoking an LLM. It does NOT specify whether the model need be executed
    remotely or locally (if applicable).
    """
    LLAMA3_2_3B = "meta-llama/Llama-3.2-3B-Instruct-Turbo"
    LLAMA3_1_8B = "meta-llama/Meta-Llama-3.1-8B-Instruct-Turbo"
    LLAMA3_3_70B = "meta-llama/Llama-3.3-70B-Instruct-Turbo"
    LLAMA3_2_90B_V = "meta-llama/Llama-3.2-90B-Vision-Instruct-Turbo"
    MIXTRAL = "mistralai/Mixtral-8x7B-Instruct-v0.1"
    DEEPSEEK_V3 = "deepseek-ai/DeepSeek-V3"
    DEEPSEEK_R1_DISTILL_QWEN_1_5B = "deepseek-ai/DeepSeek-R1-Distill-Qwen-1.5B"
    GPT_4o = "gpt-4o-2024-08-06"
    GPT_4o_MINI = "gpt-4o-mini-2024-07-18"
    TEXT_EMBEDDING_3_SMALL = "text-embedding-3-small"
    CLIP_VIT_B_32 = "clip-ViT-B-32"
    # o1 = "o1-2024-12-17"

    def __repr__(self):
        return f"{self.name}"

    def is_deepseek_model(self):
        return "deepseek" in self.value.lower()

    def is_llama_model(self):
        return "llama" in self.value.lower()

    def is_mixtral_model(self):
        return "mixtral" in self.value.lower()

    def is_clip_model(self):
        return "clip" in self.value.lower()

    def is_together_model(self):
        is_llama_model = self.is_llama_model()
        is_mixtral_model = self.is_mixtral_model()
        is_deepseek_model = self.is_deepseek_model()
        is_clip_model = self.is_clip_model()
        return is_llama_model or is_mixtral_model or is_deepseek_model or is_clip_model

    def is_gpt_4o_model(self):
        return "gpt-4o" in self.value.lower()

    def is_o1_model(self):
        return "o1" in self.value.lower()

    def is_text_embedding_model(self):
        return "text-embedding" in self.value.lower()

    def is_openai_model(self):
        is_gpt4_model = self.is_gpt_4o_model()
        is_o1_model = self.is_o1_model()
        is_text_embedding_model = self.is_text_embedding_model()
        return is_gpt4_model or is_o1_model or is_text_embedding_model

    def is_vision_model(self):
        vision_models = [
            "meta-llama/Llama-3.2-90B-Vision-Instruct-Turbo",
            "gpt-4o-2024-08-06",
            "gpt-4o-mini-2024-07-18",
            "o1-2024-12-17",
        ]
        return self.value in vision_models

    def is_embedding_model(self):
        is_clip_model = self.is_clip_model()
        is_text_embedding_model = self.is_text_embedding_model()
        return is_clip_model or is_text_embedding_model

class APIClient(str, Enum):
    """
    APIClient describes the API client to be used when invoking an LLM.
    """

    OPENAI = "openai"
    TOGETHER = "together"

class PromptStrategy(str, Enum):
    """
    PromptStrategy describes the prompting technique to be used by a Generator when
    performing some task with a specified Model.
    """

    # Chain-of-Thought Boolean Prompt Strategies
    COT_BOOL = "chain-of-thought-bool"
    # COT_BOOL_CRITIC = "chain-of-thought-bool-critic"
    # COT_BOOL_REFINE = "chain-of-thought-bool-refine"

    # Chain-of-Thought Boolean with Image Prompt Strategies
    COT_BOOL_IMAGE = "chain-of-thought-bool-image"
    # COT_BOOL_IMAGE_CRITIC = "chain-of-thought-bool-image-critic"
    # COT_BOOL_IMAGE_REFINE = "chain-of-thought-bool-image-refine"

    # Chain-of-Thought Question Answering Prompt Strategies
    COT_QA = "chain-of-thought-question"
    COT_QA_CRITIC = "chain-of-thought-question-critic"
    COT_QA_REFINE = "chain-of-thought-question-refine"

    # Chain-of-Thought Question with Image Prompt Strategies
    COT_QA_IMAGE = "chain-of-thought-question-image"
    COT_QA_IMAGE_CRITIC = "chain-of-thought-question-critic-image"
    COT_QA_IMAGE_REFINE = "chain-of-thought-question-refine-image"

    # Mixture-of-Agents Prompt Strategies
    COT_MOA_PROPOSER = "chain-of-thought-mixture-of-agents-proposer"
    COT_MOA_PROPOSER_IMAGE = "chain-of-thought-mixture-of-agents-proposer-image"
    COT_MOA_AGG = "chain-of-thought-mixture-of-agents-aggregation"

    # Split Convert Prompt Strategies
    SPLIT_PROPOSER = "split-proposer"
    SPLIT_MERGER = "split-merger"

    def is_image_prompt(self):
        return "image" in self.value

    def is_bool_prompt(self):
        return "bool" in self.value

    def is_convert_prompt(self):
        return "bool" not in self.value

    def is_critic_prompt(self):
        return "critic" in self.value

    def is_refine_prompt(self):
        return "refine" in self.value

    def is_moa_proposer_prompt(self):
        return "mixture-of-agents-proposer" in self.value

    def is_moa_aggregator_prompt(self):
        return "mixture-of-agents-aggregation" in self.value

    def is_split_proposer_prompt(self):
        return "split-proposer" in self.value

    def is_split_merger_prompt(self):
        return "split-merger" in self.value

class AggFunc(str, Enum):
    COUNT = "count"
    AVERAGE = "average"


class Cardinality(str, Enum):
    ONE_TO_ONE = "one-to-one"
    ONE_TO_MANY = "one-to-many"

    @classmethod
    def _missing_(cls, value):
        if value:
            normalized_value = "".join([x for x in value if x.isalpha()]).lower()
            for member in cls:
                normalized_member = "".join([x for x in member if x.isalpha()]).lower()
                if normalized_member == normalized_value:
                    return member
        return cls.ONE_TO_ONE


class PickOutputStrategy(str, Enum):
    CHAMPION = "champion"
    ENSEMBLE = "ensemble"


IMAGE_EXTENSIONS = [".jpg", ".jpeg", ".png", ".gif", ".bmp", ".tiff"]
PDF_EXTENSIONS = [".pdf"]
XLS_EXTENSIONS = [".xls", ".xlsx"]
HTML_EXTENSIONS = [".html", ".htm"]

# the number of seconds the parallel execution will sleep for while waiting for futures to complete
PARALLEL_EXECUTION_SLEEP_INTERVAL_SECS = 0.3

# default PDF parser
DEFAULT_PDF_PROCESSOR = "pypdf"

# character limit for various IDs
MAX_ID_CHARS = 10

# maximum number of rows to display in a table
MAX_ROWS = 5

# maximum number of rows to parse from an HTML
MAX_HTML_ROWS = 10000


def log_attempt_number(retry_state):
    """return the result of the last call attempt"""
    print(f"Retrying: {retry_state.attempt_number}...")


# Palimpzest root directory
PZ_DIR = os.path.join(os.path.expanduser("~"), ".palimpzest")

# Assume 500 MB/sec for local SSD scan time
LOCAL_SCAN_TIME_PER_KB = 1 / (float(500) * 1024)

# Assume 30 GB/sec for sequential access of memory
MEMORY_SCAN_TIME_PER_KB = 1 / (float(30) * 1024 * 1024)

# Assume 1 KB per record
NAIVE_BYTES_PER_RECORD = 1024

# Rough conversion from # of characters --> # of tokens; assumes 1 token ~= 4 chars
TOKENS_PER_CHARACTER = 0.25

# Rough estimate of the number of tokens the context is allowed to take up for MIXTRAL and LLAMA3 models
MIXTRAL_LLAMA_CONTEXT_TOKENS_LIMIT = 6000

# a naive estimate for the input record size
NAIVE_EST_SOURCE_RECORD_SIZE_IN_BYTES = 1_000_000

# a naive estimate for filter selectivity
NAIVE_EST_FILTER_SELECTIVITY = 0.5

# a naive estimate for the number of input tokens processed per record
NAIVE_EST_NUM_INPUT_TOKENS = 1000

# a naive estimate for the number of output tokens processed per record
NAIVE_EST_NUM_OUTPUT_TOKENS = 100

# a naive estimate for the number of groups returned by a group by
NAIVE_EST_NUM_GROUPS = 3

# a naive estimate for the factor of increase (loosely termed "selectivity") for one-to-many cardinality operations
NAIVE_EST_ONE_TO_MANY_SELECTIVITY = 2

# a naive estimate of the time it takes to extract the latex for an equation from an image file using Skema
NAIVE_IMAGE_TO_EQUATION_LATEX_TIME_PER_RECORD = 10.0

# a naive estimate of the time it takes to extract the text from a PDF using a PDF processor
NAIVE_PDF_PROCESSOR_TIME_PER_RECORD = 10.0

# Whether or not to log LLM outputs
LOG_LLM_OUTPUT = False


#### MODEL PERFORMANCE & COST METRICS ####
# Overall model quality is computed using MMLU-Pro; multi-modal models currently use the same score for vision
# - in the future we should split quality for vision vs. multi-modal vs. text
# - code quality was computed using HumanEval, but that benchmark is too easy and should be replaced.
# - https://huggingface.co/spaces/TIGER-Lab/MMLU-Pro
#
# Cost is presented in terms of USD / token for input tokens and USD / token for
# generated tokens.
#
# Time is presented in seconds per output token. I grabbed some semi-recent estimates
# from the internet for this quick POC, but we can and should do more to model these
# values more precisely:
# - https://artificialanalysis.ai/models/llama-3-1-instruct-8b
#
LLAMA3_2_3B_INSTRUCT_MODEL_CARD = {
    ##### Cost in USD #####
    "usd_per_input_token": 0.06 / 1e6,
    "usd_per_output_token": 0.06 / 1e6,
    ##### Time #####
    "seconds_per_output_token": 0.0064,
    ##### Agg. Benchmark #####
    "overall": 36.50, # https://huggingface.co/meta-llama/Llama-3.2-3B-Instruct/discussions/13
    ##### Code #####
    "code": 0.0,
}
LLAMA3_1_8B_INSTRUCT_MODEL_CARD = {
    ##### Cost in USD #####
    "usd_per_input_token": 0.18 / 1e6,
    "usd_per_output_token": 0.18 / 1e6,
    ##### Time #####
    "seconds_per_output_token": 0.0059,
    ##### Agg. Benchmark #####
    "overall": 44.25,
    ##### Code #####
    "code": 72.6,
}
LLAMA3_3_70B_INSTRUCT_MODEL_CARD = {
    ##### Cost in USD #####
    "usd_per_input_token": 0.88 / 1e6,
    "usd_per_output_token": 0.88 / 1e6,
    ##### Time #####
    "seconds_per_output_token": 0.0139,
    ##### Agg. Benchmark #####
    "overall": 65.92,
    ##### Code #####
    "code": 88.4,
}
LLAMA3_2_90B_V_MODEL_CARD = {
    ##### Cost in USD #####
    "usd_per_input_token": 1.2 / 1e6,
    "usd_per_output_token": 1.2 / 1e6,
    ##### Time #####
    "seconds_per_output_token": 0.0222,
    ##### Agg. Benchmark #####
    "overall": 65.00, # set to be slightly higher than gpt-4o-mini
}
MIXTRAL_8X_7B_MODEL_CARD = {
    ##### Cost in USD #####
    "usd_per_input_token": 0.6 / 1e6,
    "usd_per_output_token": 0.6 / 1e6,
    ##### Time #####
    "seconds_per_output_token": 0.0112,
    ##### Agg. Benchmark #####
    "overall": 43.27,
    ##### Code #####
    "code": 40.0,
}
DEEPSEEK_V3_MODEL_CARD = {
    ##### Cost in USD #####
    "usd_per_input_token": 1.25 / 1E6,
    "usd_per_output_token": 1.25 / 1E6,
    ##### Time #####
    "seconds_per_output_token": 0.0769,
    ##### Agg. Benchmark #####
    "overall": 75.87,
    ##### Code #####
    "code": 92.0,
}
DEEPSEEK_R1_DISTILL_QWEN_1_5B_MODEL_CARD = {
    ##### Cost in USD #####
    "usd_per_input_token": 0.18 / 1E6,
    "usd_per_output_token": 0.18 / 1E6,
    ##### Time #####
    "seconds_per_output_token": 0.0026,
    ##### Agg. Benchmark #####
    "overall": 39.90, # https://www.reddit.com/r/LocalLLaMA/comments/1iserf9/deepseek_r1_distilled_models_mmlu_pro_benchmarks/
    ##### Code #####
    "code": 0.0,
}
GPT_4o_MODEL_CARD = {
    # NOTE: it is unclear if the same ($ / token) costs can be applied for vision, or if we have to calculate this ourselves
    ##### Cost in USD #####
    "usd_per_input_token": 2.5 / 1e6,
    "usd_per_output_token": 10.0 / 1e6,
    ##### Time #####
    "seconds_per_output_token": 0.0079,
    ##### Agg. Benchmark #####
    "overall": 74.68,
    ##### Code #####
    "code": 90.0,
}
GPT_4o_MINI_MODEL_CARD = {
    # NOTE: it is unclear if the same ($ / token) costs can be applied for vision, or if we have to calculate this ourselves
    ##### Cost in USD #####
    "usd_per_input_token": 0.15 / 1e6,
    "usd_per_output_token": 0.6 / 1e6,
    ##### Time #####
    "seconds_per_output_token": 0.0098,
    ##### Agg. Benchmark #####
    "overall": 63.09,
    ##### Code #####
    "code": 86.0,
}
o1_MODEL_CARD = {  # noqa: N816
    # NOTE: it is unclear if the same ($ / token) costs can be applied for vision, or if we have to calculate this ourselves
    ##### Cost in USD #####
    "usd_per_input_token": 15 / 1e6,
    "usd_per_output_token": 60 / 1e6,
    ##### Time #####
    "seconds_per_output_token": 0.0110,
    ##### Agg. Benchmark #####
    "overall": 89.30,
    ##### Code #####
    "code": 92.3, # NOTE: just copying MMLU score for now
}
TEXT_EMBEDDING_3_SMALL_MODEL_CARD = {
    ##### Cost in USD #####
    "usd_per_input_token": 0.02 / 1e6,
    "usd_per_output_token": None,
    ##### Time #####
    "seconds_per_output_token": 0.0098,  # NOTE: just copying GPT_4o_MINI_MODEL_CARD for now
    ##### Agg. Benchmark #####
    "overall": 63.09,  # NOTE: just copying GPT_4o_MINI_MODEL_CARD for now
}
CLIP_VIT_B_32_MODEL_CARD = {
    ##### Cost in USD #####
    "usd_per_input_token": 0.00,
    "usd_per_output_token": None,
    ##### Time #####
    "seconds_per_output_token": 0.0098,  # NOTE: just copying TEXT_EMBEDDING_3_SMALL_MODEL_CARD for now
    ##### Agg. Benchmark #####
    "overall": 63.3,  # NOTE: ImageNet top-1 accuracy
}


MODEL_CARDS = {
    Model.LLAMA3_2_3B.value: LLAMA3_2_3B_INSTRUCT_MODEL_CARD,
    Model.LLAMA3_1_8B.value: LLAMA3_1_8B_INSTRUCT_MODEL_CARD,
    Model.LLAMA3_3_70B.value: LLAMA3_3_70B_INSTRUCT_MODEL_CARD,
    Model.LLAMA3_2_90B_V.value: LLAMA3_2_90B_V_MODEL_CARD,
    Model.DEEPSEEK_V3.value: DEEPSEEK_V3_MODEL_CARD,
    Model.DEEPSEEK_R1_DISTILL_QWEN_1_5B.value: DEEPSEEK_R1_DISTILL_QWEN_1_5B_MODEL_CARD,
    Model.MIXTRAL.value: MIXTRAL_8X_7B_MODEL_CARD,
    Model.GPT_4o.value: GPT_4o_MODEL_CARD,
    Model.GPT_4o_MINI.value: GPT_4o_MINI_MODEL_CARD,
    # Model.o1.value: o1_MODEL_CARD,
    Model.TEXT_EMBEDDING_3_SMALL.value: TEXT_EMBEDDING_3_SMALL_MODEL_CARD,
    Model.CLIP_VIT_B_32.value: CLIP_VIT_B_32_MODEL_CARD,
}


###### DEPRECATED ######
# # NOTE: seconds_per_output_token is based on `gpt-3.5-turbo-1106`
# GPT_3_5_MODEL_CARD = {
#     ##### Cost in USD #####
#     "usd_per_input_token": 0.5 / 1E6,
#     "usd_per_output_token": 1.5 / 1E6,
#     ##### Time #####
#     "seconds_per_output_token": 0.0065,
#     ##### Agg. Benchmark #####
#     "overall": 70.0, # 5-shot
#     ##### Commonsense Reasoning #####
#     "reasoning": 84.1,
#     ### "HellaSwag": 85.5,^  # 10-shot
#     ### "WinoGrande": 81.6,^ # 5-shot
#     ### "Arc-e": 85.2,^      # 25-shot
#     ##### World Knowledge #####
#     ##### Reading Comprehension #####
#     ### "DROP": 64.1, # 3-shot
#     ##### Code #####
#     "code": 48.1,
#     ### "HumanEval": 48.1,^ # 0-shot
#     ##### Math #####
#     "math": 57.1,
#     ### "GSM8K": 57.1,^     # 5-shot
# }
# # NOTE: the seconds_per_output_token was computed based on a slightly different model ('gpt-4-1106-preview')
# #       and the benchmark statistics were computed based on the GPT-4 Technical Report; these might be
# #       slightly innacurate compared to the real numbers for gpt-4-0125-preview, but we'll use them until
# #       we have something better. (The cost metrics are accurate).
# GPT_4_MODEL_CARD = {
#     ##### Cost in USD #####
#     "usd_per_input_token": 10 / 1E6,
#     "usd_per_output_token": 30 / 1E6,
#     ##### Time #####
#     "seconds_per_output_token": 0.018,
#     ##### Agg. Benchmark #####
#     "overall": 86.4, # 5-shot
#     ##### Commonsense Reasoning #####
#     "reasoning": 93.0,
#     ### "HellaSwag": 95.3,^  # 10-shot
#     ### "WinoGrande": 87.5,^ # 5-shot
#     ### "Arc-e": 96.3,^      # 25-shot
#     ##### World Knowledge #####
#     ##### Reading Comprehension #####
#     ### "DROP": 80.9, # 3-shot
#     ##### Code #####
#     "code": 67.0,
#     ### "HumanEval": 67.0,^ # 0-shot
#     ##### Math #####
#     "math": 92.0,
#     ### "GSM8K": 92.0,^     # 5-shot
# }

# # TODO: use cost info in here: https://platform.openai.com/docs/guides/vision/calculating-costs
# GPT_4V_MODEL_CARD = {
#     ##### Cost in USD #####
#     "usd_per_input_token": 10 / 1E6,
#     "usd_per_output_token": 30 / 1E6,
#     ##### Time #####
#     "seconds_per_output_token": 0.042 / 10.0, # TODO: / 10.0 is a hack; need to figure out why time estimates are so off
#     ##### Agg. Benchmark #####
#     "overall": 86.4,
# }


# GEMINI_1_MODEL_CARD = {
#     ##### Cost in USD #####
#     "usd_per_input_token": 125 / 1E8, # Gemini is free but rate limited for now. Pricing will be updated
#     "usd_per_output_token": 375 / 1E9,
#     ##### Time #####
#     "seconds_per_output_token": 0.042 / 10.0, # TODO:
#     ##### Agg. Benchmark #####
#     "overall": 65.0, # 90.0 TODO: we are using the free version of Gemini which is substantially worse than its paid version; I'm manually revising it's quality below that of Mixtral
#     ##### Commonsense Reasoning #####
#     "reasoning": 80.0, # 87.8, TODO: see note above on overall
#     # "HellaSwag": 87.8,  # 10-shot
#     ##### World Knowledge #####
#     ##### Reading Comprehension #####
#     # "DROP": 82.4, # Variable shots ?
#     ##### Code #####
#     "code": 74.4,
#     # "HumanEval": 74.4, # 0-shot (IT)*
#     # "Natural2Code": 74.9, # 0-shot
#     ##### Math #####
#     "math": 94.4,
#     # "GSM8K": 94.4,     # maj1@32
#     # "MATH": 53.2,      # 4-shot
# }

# GEMINI_1V_MODEL_CARD = {
#     ##### Cost in USD #####
#     "usd_per_input_token": 25 / 1E6,  # Gemini is free but rate limited for now. Pricing will be updated
#     "usd_per_output_token": 375 / 1E9,
#     ##### Time #####
#     "seconds_per_output_token": 0.042, # / 10.0, # TODO:
#     ##### Agg. Benchmark #####
#     "overall": 65.0, # 90.0, TODO: see note above in Gemini_1 model card
#     ##### Commonsense Reasoning #####
#     "reasoning": 80.0, # 87.8, TODO: see note above in Gemini_1 model card
#     # "HellaSwag": 87.8,  # 10-shot
#     ##### World Knowledge #####
#     ##### Reading Comprehension #####
#     # "DROP": 82.4, # Variable shots ?
#     ##### Code #####
#     "code": 74.4,
#     # "HumanEval": 74.4, # 0-shot (IT)*
#     # "Natural2Code": 74.9, # 0-shot
#     ##### Math #####
#     "math": 94.4,
#     # "GSM8K": 94.4,     # maj1@32
#     # "MATH": 53.2,      # 4-shot
# }



from __future__ import annotations

import time
from abc import ABC, abstractmethod
from typing import Callable

from palimpzest.constants import (
    MODEL_CARDS,
    NAIVE_EST_NUM_INPUT_TOKENS,
    NAIVE_EST_NUM_OUTPUT_TOKENS,
    NAIVE_EST_ONE_TO_MANY_SELECTIVITY,
    Cardinality,
    Model,
    PromptStrategy,
)
from palimpzest.core.data.dataclasses import GenerationStats, OperatorCostEstimates, RecordOpStats
from palimpzest.core.elements.records import DataRecord, DataRecordSet
from palimpzest.core.lib.fields import Field
from palimpzest.query.generators.generators import generator_factory
from palimpzest.query.operators.physical import PhysicalOperator
from palimpzest.utils.model_helpers import get_vision_models


class ConvertOp(PhysicalOperator, ABC):
    def __init__(
        self,
        cardinality: Cardinality = Cardinality.ONE_TO_ONE,
        udf: Callable | None = None,
        desc: str | None = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.cardinality = cardinality
        self.udf = udf
        self.desc = desc

    def get_id_params(self):
        id_params = super().get_id_params()
        id_params = {
            "cardinality": self.cardinality.value,
            "udf": self.udf,
            **id_params,
        }

        return id_params

    def get_op_params(self):
        op_params = super().get_op_params()
        op_params = {"cardinality": self.cardinality, "udf": self.udf, "desc": self.desc, **op_params}

        return op_params

    def _create_data_records_from_field_answers(
        self,
        field_answers: dict[str, list],
        candidate: DataRecord,
    ) -> list[DataRecord]:
        """
        Given a mapping from each field to its (list of) generated value(s), we construct the corresponding
        list of output DataRecords.
        """
        # get the number of records generated; for some convert operations it is possible for fields to
        # have different lengths of generated values, so we take the maximum length of any field's values
        # to be the number of records generated
        n_records = max([len(lst) for lst in field_answers.values()])
        successful_convert = n_records > 0

        drs = []
        for idx in range(max(n_records, 1)):
            # initialize record with the correct output schema, parent record, and cardinality idx
            dr = DataRecord.from_parent(self.output_schema, parent_record=candidate, cardinality_idx=idx)

            # copy all fields from the input record
            # NOTE: this means that records processed by PZ converts will inherit all pre-computed fields
            #       in an incremental fashion; this is a design choice which may be revisited in the future
            for field in candidate.get_field_names():
                setattr(dr, field, getattr(candidate, field))

            # get input field names and output field names
            input_fields = self.input_schema.field_names()
            output_fields = self.output_schema.field_names()

            # parse newly generated fields from the field_answers dictionary for this field; if the list
            # of generated values is shorter than the number of records, we fill in with None
            for field in output_fields:
                if field not in input_fields:
                    value = field_answers[field][idx] if idx < len(field_answers[field]) else None
                    setattr(dr, field, value)

            # append data record to list of output data records
            drs.append(dr)

        return drs, successful_convert

    def _create_record_set(
        self,
        records: list[DataRecord],
        field_names: list[str],
        generation_stats: GenerationStats,
        total_time: float,
        successful_convert: bool,
    ) -> DataRecordSet:
        """
        Construct list of RecordOpStats objects (one for each DataRecord).
        """
        # amortize the generation stats across all generated records
        per_record_stats = generation_stats / len(records)
        time_per_record = total_time / len(records)

        # create the RecordOpStats objects for each output record
        record_op_stats_lst = [
            RecordOpStats(
                record_id=dr.id,
                record_parent_id=dr.parent_id,
                record_source_idx=dr.source_idx,
                record_state=dr.to_dict(include_bytes=False),
                full_op_id=self.get_full_op_id(),
                logical_op_id=self.logical_op_id,
                op_name=self.op_name(),
                time_per_record=time_per_record,
                cost_per_record=per_record_stats.cost_per_record,
                model_name=self.get_model_name(),
                answer={field_name: getattr(dr, field_name) for field_name in field_names},
                input_fields=self.input_schema.field_names(),
                generated_fields=field_names,
                total_input_tokens=per_record_stats.total_input_tokens,
                total_output_tokens=per_record_stats.total_output_tokens,
                total_input_cost=per_record_stats.total_input_cost,
                total_output_cost=per_record_stats.total_output_cost,
                llm_call_duration_secs=per_record_stats.llm_call_duration_secs,
                fn_call_duration_secs=per_record_stats.fn_call_duration_secs,
                total_llm_calls=per_record_stats.total_llm_calls,
                total_embedding_llm_calls=per_record_stats.total_embedding_llm_calls,
                failed_convert=(not successful_convert),
                image_operation=self.is_image_conversion(),
                op_details={k: str(v) for k, v in self.get_id_params().items()},
            )
            for dr in records
        ]

        # create and return the DataRecordSet
        return DataRecordSet(records, record_op_stats_lst)

    @abstractmethod
    def is_image_conversion(self) -> bool:
        """Return True if the convert operation processes an image, False otherwise."""
        pass

    @abstractmethod
    def convert(self, candidate: DataRecord, fields: dict[str, Field]) -> tuple[dict[str, list], GenerationStats]:
        """
        This abstract method will be implemented by subclasses of ConvertOp to process the input DataRecord
        and generate the value(s) for each of the specified fields. If the convert operator is a one-to-many
        convert, then each field will have a corresponding list of output values. The dictionary mapping each
        generated field to its (list of) value(s) is returned along with the GenerationStats object.

        For example, if the input DataRecord (i.e. `candidate`) contains the contents of a scientific paper,
        and the convert operation is supposed to extract the name and affiliation of each author into its own
        DataRecord, then the output could be:

        ({"author": ["Jane Smith", "John Doe"], "affiliation": ["MIT", "Stanford University"]}, GenerationStats(...))

        Even if the convert operation is a one-to-one convert (i.e. it always generates one output DataRecord
        for each input DataRecord), the output should still map each field to a singleton list containing its value.

        A post-condition of this method is that every field in `fields` must be present in the output dictionary.
        If there is an error in generating a field, then the value for that field must be None.
        """
        pass

    def __call__(self, candidate: DataRecord) -> DataRecordSet:
        """
        This method converts an input DataRecord into an output DataRecordSet. The output DataRecordSet contains the
        DataRecord(s) output by the operator's convert() method and their corresponding RecordOpStats objects.
        Some subclasses may override this __call__method to implement their own custom logic.
        """
        start_time = time.time()

        # get fields to generate with this convert
        fields_to_generate = self.get_fields_to_generate(candidate)

        # execute the convert
        field_answers: dict[str, list]
        fields = {field: field_type for field, field_type in self.output_schema.field_map().items() if field in fields_to_generate}
        field_answers, generation_stats = self.convert(candidate=candidate, fields=fields)
        assert all([field in field_answers for field in fields_to_generate]), "Not all fields were generated!"

        # replace any None values with an empty list; subclasses may override __call__ to change this behavior
        field_answers = {field: [] if answers is None else answers for field, answers in field_answers.items()}

        # transform the mapping from fields to answers into a (list of) DataRecord(s)
        drs, successful_convert = self._create_data_records_from_field_answers(field_answers, candidate)

        # construct and return DataRecordSet
        record_set = self._create_record_set(
            records=drs,
            field_names=fields_to_generate,
            generation_stats=generation_stats,
            total_time=time.time() - start_time,
            successful_convert=successful_convert,
        )

        return record_set


class NonLLMConvert(ConvertOp):
    def __str__(self):
        op = super().__str__()
        op += f"    UDF: {self.udf.__name__}\n"
        return op

    def is_image_conversion(self) -> bool:
        # NOTE: even if the UDF is processing an image, we do not consider this an image conversion
        # (the output of this function will be used by the CostModel in a way which does not apply to UDFs)
        return False

    def naive_cost_estimates(self, source_op_cost_estimates: OperatorCostEstimates) -> OperatorCostEstimates:
        """
        Compute naive cost estimates for the NonLLMConvert operation. These estimates assume
        that the UDF convert (1) has no cost and (2) has perfect quality.
        """
        # estimate cardinality and selectivity given the "cardinality" set by the user
        selectivity = 1.0 if self.cardinality == Cardinality.ONE_TO_ONE else NAIVE_EST_ONE_TO_MANY_SELECTIVITY
        cardinality = selectivity * source_op_cost_estimates.cardinality

        # estimate 1 ms single-threaded execution for udf function
        time_per_record = 0.001

        # assume filter fn has perfect quality
        return OperatorCostEstimates(
            cardinality=cardinality,
            time_per_record=time_per_record,
            cost_per_record=0.0,
            quality=1.0,
        )

    def convert(self, candidate: DataRecord, fields: dict[str, Field]) -> tuple[dict[str, list], GenerationStats]:
        # apply UDF to input record
        start_time = time.time()
        field_answers = {}
        try:
            # execute the UDF function
            answer = self.udf(candidate.to_dict())

            if self.cardinality == Cardinality.ONE_TO_ONE:
                # answer should be a dictionary
                assert isinstance(answer, dict), (
                    "UDF must return a dictionary mapping each generated field to its value for one-to-one converts"
                )

                # wrap each answer in a list
                field_answers = {field_name: [answer[field_name]] for field_name in fields}

            else:
                assert isinstance(answer, list), "UDF must return a list of dictionaries for one-to-many converts"
                field_answers = {field_name: [] for field_name in fields}
                for answer_dict in answer:
                    assert isinstance(answer_dict, dict), "Each element of list returned by UDF must be a dictionary"
                    for field_name in fields:
                        field_answers[field_name].append(answer_dict.get(field_name, None))

            if self.verbose:
                print(f"{self.udf.__name__}:\n{answer}")

        except Exception as e:
            print(f"Error invoking user-defined function for convert: {e}")
            raise e

        # create generation stats object containing the time spent executing the UDF function
        generation_stats = GenerationStats(fn_call_duration_secs=time.time() - start_time)

        return field_answers, generation_stats


class LLMConvert(ConvertOp):
    """
    This is the base class for convert operations which use an LLM to generate the output fields.
    """

    def __init__(
        self,
        model: Model,
        prompt_strategy: PromptStrategy = PromptStrategy.COT_QA,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.model = model
        self.prompt_strategy = prompt_strategy
        if model is not None:
            self.generator = generator_factory(model, prompt_strategy, self.cardinality, self.verbose)

    def __str__(self):
        op = super().__str__()
        op += f"    Prompt Strategy: {self.prompt_strategy}\n"
        return op

    def get_id_params(self):
        id_params = super().get_id_params()
        id_params = {
            "model": None if self.model is None else self.model.value,
            "prompt_strategy": None if self.prompt_strategy is None else self.prompt_strategy.value,
            **id_params,
        }

        return id_params

    def get_op_params(self):
        op_params = super().get_op_params()
        op_params = {
            "model": self.model,
            "prompt_strategy": self.prompt_strategy,
            **op_params,
        }

        return op_params

    def get_model_name(self):
        return None if self.model is None else self.model.value

    def is_image_conversion(self) -> bool:
        return self.model in get_vision_models()

    def naive_cost_estimates(self, source_op_cost_estimates: OperatorCostEstimates) -> OperatorCostEstimates:
        """
        Compute naive cost estimates for the LLMConvert operation. Implicitly, these estimates
        assume the use of a single LLM call for each input record. Child classes of LLMConvert
        may call this function through super() and adjust these estimates as needed (or they can
        completely override this function).
        """
        # estimate number of input and output tokens from source
        est_num_input_tokens = NAIVE_EST_NUM_INPUT_TOKENS
        est_num_output_tokens = NAIVE_EST_NUM_OUTPUT_TOKENS

        # get est. of conversion time per record from model card;
        # NOTE: model will only be None for code synthesis, which uses GPT-3.5 as fallback
        model_name = self.model.value if getattr(self, "model", None) is not None else Model.GPT_4o_MINI.value
        model_conversion_time_per_record = MODEL_CARDS[model_name]["seconds_per_output_token"] * est_num_output_tokens

        # get est. of conversion cost (in USD) per record from model card
        model_conversion_usd_per_record = (
            MODEL_CARDS[model_name]["usd_per_input_token"] * est_num_input_tokens
            + MODEL_CARDS[model_name]["usd_per_output_token"] * est_num_output_tokens
        )

        # estimate cardinality and selectivity given the "cardinality" set by the user
        selectivity = 1.0 if self.cardinality == Cardinality.ONE_TO_ONE else NAIVE_EST_ONE_TO_MANY_SELECTIVITY
        cardinality = selectivity * source_op_cost_estimates.cardinality

        # estimate quality of output based on the strength of the model being used
        quality = (MODEL_CARDS[model_name]["overall"] / 100.0) * source_op_cost_estimates.quality

        return OperatorCostEstimates(
            cardinality=cardinality,
            time_per_record=model_conversion_time_per_record,
            cost_per_record=model_conversion_usd_per_record,
            quality=quality,
        )


class LLMConvertBonded(LLMConvert):

    def convert(self, candidate: DataRecord, fields: dict[str, Field]) -> tuple[dict[str, list], GenerationStats]:
        # get the set of input fields to use for the convert operation
        input_fields = self.get_input_fields()

        # construct kwargs for generation
        gen_kwargs = {"project_cols": input_fields, "output_schema": self.output_schema}

        # generate outputs for all fields in a single query
        field_answers, _, generation_stats, _ = self.generator(candidate, fields, **gen_kwargs)

        # if there was an error for any field, execute a conventional query on that field
        if len(field_answers) > 1:
            for field_name, answers in field_answers.items():
                if answers is None:
                    single_field_answers, _, single_field_stats, _ = self.generator(candidate, {field_name: fields[field_name]}, **gen_kwargs)
                    field_answers.update(single_field_answers)
                    generation_stats += single_field_stats

        return field_answers, generation_stats



import os

from palimpzest.constants import Model


def get_vision_models() -> list[Model]:
    """
    Return the set of vision models which the system has access to based on the set of environment variables.
    """
    models = []
    if os.getenv("OPENAI_API_KEY") is not None:
        openai_vision_models = [
            model for model in Model
            if model.is_openai_model() and model.is_vision_model()
        ]
        models.extend(openai_vision_models)

    if os.getenv("TOGETHER_API_KEY") is not None:
        together_vision_models = [
            model for model in Model
            if model.is_together_model() and model.is_vision_model()
        ]
        models.extend(together_vision_models)

    return models


def get_models(include_vision: bool = False) -> list[Model]:
    """
    Return the set of models which the system has access to based on the set environment variables.
    """
    models = []
    if os.getenv("OPENAI_API_KEY") is not None:
        openai_models = [model for model in Model if model.is_openai_model()]
        models.extend(openai_models)

    if os.getenv("TOGETHER_API_KEY") is not None:
        together_models = [model for model in Model if model.is_together_model()]
        if not include_vision:
            together_models = [
                model for model in together_models if not model.is_vision_model()
            ]
        models.extend(together_models)

    if include_vision:
        vision_models = get_vision_models()
        models.extend(vision_models)

    return models

# The order is the priority of the model
TEXT_MODEL_PRIORITY = [
    # Model.o1,
    Model.GPT_4o,
    Model.GPT_4o_MINI,
    Model.LLAMA3_3_70B,
    Model.MIXTRAL,
    Model.DEEPSEEK_V3,
    Model.LLAMA3_2_3B,
    Model.LLAMA3_1_8B,
    Model.DEEPSEEK_R1_DISTILL_QWEN_1_5B,
]

VISION_MODEL_PRIORITY = [
    Model.GPT_4o,
    Model.GPT_4o_MINI,
    Model.LLAMA3_2_90B_V,
]
def get_champion_model(available_models, vision=False):
    # Select appropriate priority list based on task
    model_priority = VISION_MODEL_PRIORITY if vision else TEXT_MODEL_PRIORITY

    # Return first available model from priority list
    for model in model_priority:
        if model in available_models:
            return model

    # If no suitable model found, raise informative error
    task_type = "vision" if vision else "text"
    raise Exception(
        f"No {task_type} models available to create physical plans!\n"
        "You must set at least one of the following environment variables:\n"
        "[OPENAI_API_KEY, TOGETHER_API_KEY, GOOGLE_API_KEY]\n"
        f"Available models: {available_models}"
    )


def get_fallback_model(available_models, vision=False):
    return get_champion_model(available_models, vision)


def get_code_champion_model(available_models):
    # NOTE: for now, assume same champion as get_champion_model()
    return get_champion_model(available_models, vision=False)


def get_champion_model_name(available_models, vision=False):
    return get_champion_model(available_models, vision).value



from __future__ import annotations

import time
from abc import abstractmethod
from dataclasses import dataclass, field, fields
from typing import Any

import numpy as np


@dataclass
class GenerationStats:
    """
    Dataclass for storing statistics about the execution of an operator on a single record.
    """

    model_name: str | None = None

    # The raw answer as output from the generator (a list of strings, possibly of len 1)
    # raw_answers: Optional[List[str]] = field(default_factory=list)

    # the total number of input tokens processed by this operator; None if this operation did not use an LLM
    # typed as a float because GenerationStats may be amortized (i.e. divided) across a number of output records
    total_input_tokens: float = 0.0

    # the total number of output tokens processed by this operator; None if this operation did not use an LLM
    # typed as a float because GenerationStats may be amortized (i.e. divided) across a number of output records
    total_output_tokens: float = 0.0

    # the total cost of processing the input tokens; None if this operation did not use an LLM
    total_input_cost: float = 0.0

    # the total cost of processing the output tokens; None if this operation did not use an LLM
    total_output_cost: float = 0.0

    # the total cost of processing the output tokens; None if this operation did not use an LLM
    cost_per_record: float = 0.0

    # (if applicable) the time (in seconds) spent executing a call to an LLM
    llm_call_duration_secs: float = 0.0

    # (if applicable) the time (in seconds) spent executing a call to a function
    fn_call_duration_secs: float = 0.0

    # (if applicable) the total number of LLM calls made by this operator
    total_llm_calls: int = 0

    # (if applicable) the total number of embedding LLM calls made by this operator
    total_embedding_llm_calls: int = 0

    def __iadd__(self, other: GenerationStats) -> GenerationStats:
        # self.raw_answers.extend(other.raw_answers)
        for dataclass_field in [
            "total_input_tokens",
            "total_output_tokens",
            "total_input_cost",
            "total_output_cost",
            "cost_per_record",
            "llm_call_duration_secs",
            "fn_call_duration_secs",
            "total_llm_calls",
            "total_embedding_llm_calls",
        ]:
            setattr(self, dataclass_field, getattr(self, dataclass_field) + getattr(other, dataclass_field))
        return self

    def __add__(self, other: GenerationStats) -> GenerationStats:
        dct = {
            field: getattr(self, field) + getattr(other, field)
            for field in [
                "total_input_tokens",
                "total_output_tokens",
                "total_input_cost",
                "total_output_cost",
                "llm_call_duration_secs",
                "fn_call_duration_secs",
                "cost_per_record",
                "total_llm_calls",
                "total_embedding_llm_calls",
            ]
        }
        # dct['raw_answers'] = self.raw_answers + other.raw_answers
        dct["model_name"] = self.model_name
        return GenerationStats(**dct)

    # Do the same as iadd and add but with division operator
    def __itruediv__(self, quotient: float) -> GenerationStats:
        if quotient == 0:
            raise ZeroDivisionError("Cannot divide by zero")
        if isinstance(quotient, int):
            quotient = float(quotient)
        for dataclass_field in [
            "total_input_tokens",
            "total_output_tokens",
            "total_input_cost",
            "total_output_cost",
            "cost_per_record",
            "llm_call_duration_secs",
            "fn_call_duration_secs",
            "total_llm_calls",
            "total_embedding_llm_calls",
        ]:
            setattr(self, dataclass_field, getattr(self, dataclass_field) / quotient)
        return self

    def __truediv__(self, quotient: float) -> GenerationStats:
        if quotient == 0:
            raise ZeroDivisionError("Cannot divide by zero")
        if isinstance(quotient, int):
            quotient = float(quotient)
        dct = {
            field: getattr(self, field) / quotient
            for field in [
                "total_input_tokens",
                "total_output_tokens",
                "total_input_cost",
                "total_output_cost",
                "llm_call_duration_secs",
                "fn_call_duration_secs",
                "total_llm_calls",
                "total_embedding_llm_calls",
                "cost_per_record",
            ]
        }
        dct["model_name"] = self.model_name
        return GenerationStats(**dct)

    def __radd__(self, other: int) -> GenerationStats:
        assert not isinstance(other, GenerationStats), "This should not be called with a GenerationStats object"
        return self


@dataclass
class RecordOpStats:
    """
    Dataclass for storing statistics about the execution of an operator on a single record.
    """

    ##### REQUIRED FIELDS #####
    # record id; an identifier for this record
    record_id: str

    # identifier for the parent of this record
    record_parent_id: str

    # idenifier for the source idx of this record
    record_source_idx: str

    # a dictionary with the record state after being processed by the operator
    record_state: dict[str, Any]

    # operation id; an identifier for this operation's physical op id
    full_op_id: str

    # logical operation id; the logical op id for this physical op
    logical_op_id: str

    # operation name
    op_name: str

    # the time spent by the data record just in this operation
    time_per_record: float

    # the cost (in dollars) to generate this record at this operation
    cost_per_record: float

    ##### NOT-OPTIONAL, BUT FILLED BY EXECUTION CLASS AFTER CONSTRUCTOR CALL #####
    # the ID of the physical operation which produced the input record for this record at this operation
    source_full_op_id: str | None = None

    # the ID of the physical plan which produced this record at this operation
    plan_id: str = ""

    ##### OPTIONAL, BUT FILLED BY COST MODEL AFTER SAMPLE DATA EXECUTION #####
    quality: float | None = None

    ##### OPTIONAL FIELDS (I.E. ONLY MANDATORY FOR CERTAIN OPERATORS) #####
    # (if applicable) the name of the model used to generate the output for this record
    model_name: str | None = None

    # (if applicable) the mapping from field-name to generated output for this record
    answer: dict[str, Any] | None = None

    # (if applicable) the mapping from field-name to generated output for this record
    # raw_answers: Optional[List[str, Any]] = field(default_factory=list)

    # (if applicable) the list of input fields for the generation for this record
    input_fields: list[str] | None = None

    # (if applicable) the list of generated fields for this record
    generated_fields: list[str] | None = None

    # the total number of input tokens processed by this operator; None if this operation did not use an LLM
    # typed as a float because GenerationStats may be amortized (i.e. divided) across a number of output records
    total_input_tokens: float = 0.0

    # the total number of output tokens processed by this operator; None if this operation did not use an LLM
    # typed as a float because GenerationStats may be amortized (i.e. divided) across a number of output records
    total_output_tokens: float = 0.0

    # the total cost of processing the input tokens; None if this operation did not use an LLM
    total_input_cost: float = 0.0

    # the total cost of processing the output tokens; None if this operation did not use an LLM
    total_output_cost: float = 0.0

    # (if applicable) the filter text (or a string representation of the filter function) applied to this record
    filter_str: str | None = None

    # the True/False result of whether this record was output by the operator or not
    # (can only be False if the operator is as Filter)
    passed_operator: bool = True

    # (if applicable) the time (in seconds) spent executing a call to an LLM
    llm_call_duration_secs: float = 0.0

    # (if applicable) the time (in seconds) spent executing a UDF or calling an external api
    fn_call_duration_secs: float = 0.0

    # (if applicable) the total number of LLM calls made by this operator
    total_llm_calls: int = 0

    # (if applicable) the total number of embedding LLM calls made by this operator
    total_embedding_llm_calls: int = 0

    # (if applicable) a boolean indicating whether this is the statistics captured from a failed convert operation
    failed_convert: bool | None = None

    # (if applicable) a boolean indicating whether this is an image convert/filter operation
    image_operation: bool | None = None

    # an OPTIONAL dictionary with more detailed information about this operation;
    op_details: dict[str, Any] = field(default_factory=dict)

    def to_json(self):
        return {field.name: getattr(self, field.name) for field in fields(self)}


@dataclass
class OperatorStats:
    """
    Dataclass for storing statistics captured within a given operator.
    """

    # the full ID of the physical operation in which these stats were collected
    full_op_id: str

    # the name of the physical operation in which these stats were collected
    op_name: str

    # the total time spent in this operation
    total_op_time: float = 0.0

    # the total cost of this operation
    total_op_cost: float = 0.0

    # a list of RecordOpStats processed by the operation
    record_op_stats_lst: list[RecordOpStats] = field(default_factory=list)

    # the full ID of the physical operator which precedes this one
    source_full_op_id: str | None = None

    # the ID of the physical plan which this operator is part of
    plan_id: str = ""

    # an OPTIONAL dictionary with more detailed information about this operation;
    op_details: dict[str, Any] = field(default_factory=dict)

    def __iadd__(self, stats: OperatorStats | RecordOpStats) -> OperatorStats:
        """
        Sum the given stats to this operator's stats. The given stats can be either:

        1. an OperatorStats object
        2. a RecordOpStats object

        NOTE: in case (1.) we assume the execution layer guarantees that `stats` is
              generated by the same operator in the same plan. Thus, we assume the
              full_op_ids, op_name, source_op_id, etc. do not need to be updated.
        """
        if isinstance(stats, OperatorStats):
            self.total_op_time += stats.total_op_time
            self.total_op_cost += stats.total_op_cost
            self.record_op_stats_lst.extend(stats.record_op_stats_lst)

        elif isinstance(stats, RecordOpStats):
            stats.source_full_op_id = self.source_full_op_id
            stats.plan_id = self.plan_id
            self.record_op_stats_lst.append(stats)
            self.total_op_time += stats.time_per_record
            self.total_op_cost += stats.cost_per_record

        else:
            raise TypeError(f"Cannot add {type(stats)} to OperatorStats")

        return self

    def to_json(self):
        return {
            "full_op_id": self.full_op_id,
            "op_name": self.op_name,
            "total_op_time": self.total_op_time,
            "total_op_cost": self.total_op_cost,
            "record_op_stats_lst": [record_op_stats.to_json() for record_op_stats in self.record_op_stats_lst],
            "op_details": self.op_details,
        }

@dataclass
class BasePlanStats:
    """
    Dataclass for storing statistics captured for an entire plan.

    This class is subclassed for tracking:
    - PlanStats: the statistics for execution of a PhysicalPlan
    - SentinelPlanStats: the statistics for execution of a SentinelPlan

    The key difference between the two subclasses is that the `operator_stats`
    field in the PlanStats maps from the physical operator ids to their corresponding
    OperatorStats objects.

    The `operator_stats` field in the SentinelPlanStats maps from a logical operator id
    to another dictionary which maps from the physical operator ids to their corresponding
    OperatorStats objects.
    """

    # id for identifying the physical plan
    plan_id: str

    # string representation of the physical plan
    plan_str: str | None = None

    # dictionary whose values are OperatorStats objects;
    # PlanStats maps {full_op_id -> OperatorStats}
    # SentinelPlanStats maps {logical_op_id -> {full_op_id -> OperatorStats}}
    operator_stats: dict = field(default_factory=dict)

    # total runtime for the plan measured from the start to the end of PhysicalPlan.execute()
    total_plan_time: float = 0.0

    # total cost for plan
    total_plan_cost: float = 0.0

    # start time for the plan execution; should be set by calling PlanStats.start()
    start_time: float | None = None

    def start(self) -> None:
        """Start the timer for this plan execution."""
        self.start_time = time.time()

    def finish(self) -> None:
        """Finish the timer for this plan execution."""
        if self.start_time is None:
            raise RuntimeError("PlanStats.start() must be called before PlanStats.finish()")
        self.total_plan_time = time.time() - self.start_time
        self.total_plan_cost = self.sum_op_costs()

    @staticmethod
    @abstractmethod
    def from_plan(plan) -> BasePlanStats:
        """
        Initialize this PlanStats object from a PhysicalPlan or SentinelPlan object.
        """
        pass

    @abstractmethod
    def sum_op_costs(self) -> float:
        """
        Sum the costs of all operators in this plan.
        """
        pass

    @abstractmethod
    def add_record_op_stats(self, record_op_stats: RecordOpStats | list[RecordOpStats]) -> None:
        """
        Add the given RecordOpStats to this plan's operator stats.
        """
        pass

    @abstractmethod
    def __iadd__(self, plan_stats: BasePlanStats) -> None:
        """
        Add the given PlanStats to this plan's operator stats.
        """
        pass

    @abstractmethod
    def __str__(self) -> str:
        """
        Return a string representation of this plan's statistics.
        """
        pass

    @abstractmethod
    def to_json(self) -> dict:
        """
        Return a JSON representation of this plan's statistics.
        """
        pass

@dataclass
class PlanStats(BasePlanStats):
    """
    Subclass of BasePlanStats which captures statistics from the execution of a single PhysicalPlan.
    """
    @staticmethod
    def from_plan(plan) -> PlanStats:
        """
        Initialize this PlanStats object from a PhysicalPlan object.
        """
        operator_stats = {}
        for op_idx, op in enumerate(plan.operators):
            full_op_id = op.get_full_op_id()
            operator_stats[full_op_id] = OperatorStats(
                full_op_id=full_op_id,
                op_name=op.op_name(),
                source_full_op_id=None if op_idx == 0 else plan.operators[op_idx - 1].get_full_op_id(),
                plan_id=plan.plan_id,
                op_details={k: str(v) for k, v in op.get_id_params().items()},
            )
    
        return PlanStats(plan_id=plan.plan_id, plan_str=str(plan), operator_stats=operator_stats)
 
    def sum_op_costs(self) -> float:
        """
        Sum the costs of all operators in this plan.
        """
        return sum([op_stats.total_op_cost for _, op_stats in self.operator_stats.items()])

    def add_record_op_stats(self, record_op_stats: RecordOpStats | list[RecordOpStats]) -> None:
        """
        Add the given RecordOpStats to this plan's operator stats.
        """
        # normalize input type to be list[RecordOpStats]
        record_op_stats_lst = record_op_stats if isinstance(record_op_stats, list) else [record_op_stats]

        # update operator stats
        for record_op_stats in record_op_stats_lst:
            full_op_id = record_op_stats.full_op_id
            if full_op_id in self.operator_stats:
                self.operator_stats[full_op_id] += record_op_stats
            else:
                raise ValueError(f"RecordOpStats with full_op_id {full_op_id} not found in PlanStats")

    def __iadd__(self, plan_stats: PlanStats) -> None:
        """
        NOTE: we assume the execution layer guarantees:
        1. these plan_stats belong to the same plan
        2. these plan_stats come from sequential (non-overlapping) executions of the same plan

        The latter criteria implies it is okay for this method to sum the plan (and operator) runtimes.
        """
        self.total_plan_time += plan_stats.total_plan_time
        self.total_plan_cost += plan_stats.total_plan_cost
        for full_op_id, op_stats in plan_stats.operator_stats.items():
            if full_op_id in self.operator_stats:
                self.operator_stats[full_op_id] += op_stats
            else:
                self.operator_stats[full_op_id] = op_stats

    def __str__(self) -> str:
        stats = f"total_plan_time={self.total_plan_time} \n"
        stats += f"total_plan_cost={self.total_plan_cost} \n"
        for idx, op_stats in enumerate(self.operator_stats.values()):
            stats += f"{idx}. {op_stats.op_name} time={op_stats.total_op_time} cost={op_stats.total_op_cost} \n"
        return stats

    def to_json(self) -> dict:
        return {
            "plan_id": self.plan_id,
            "plan_str": self.plan_str,
            "operator_stats": {full_op_id: op_stats.to_json() for full_op_id, op_stats in self.operator_stats.items()},
            "total_plan_time": self.total_plan_time,
            "total_plan_cost": self.total_plan_cost,
        }


@dataclass
class SentinelPlanStats(BasePlanStats):
    """
    Subclass of BasePlanStats which captures statistics from the execution of a single SentinelPlan.
    """
    @staticmethod
    def from_plan(plan) -> SentinelPlanStats:
        """
        Initialize this PlanStats object from a Sentinel object.
        """
        operator_stats = {}
        for op_set_idx, (logical_op_id, op_set) in enumerate(plan):
            operator_stats[logical_op_id] = {}
            for physical_op in op_set:
                full_op_id = physical_op.get_full_op_id()
                operator_stats[logical_op_id][full_op_id] = OperatorStats(
                    full_op_id=full_op_id,
                    op_name=physical_op.op_name(),
                    source_full_op_id=None if op_set_idx == 0 else plan.logical_op_ids[op_set_idx - 1],  # NOTE: this may be a reason to keep `source_op_id` instead of `source_full_op_id`
                    plan_id=plan.plan_id,
                    op_details={k: str(v) for k, v in physical_op.get_id_params().items()},
                )
    
        return SentinelPlanStats(plan_id=plan.plan_id, plan_str=str(plan), operator_stats=operator_stats)

    def sum_op_costs(self) -> float:
        """
        Sum the costs of all operators in this plan.
        """
        return sum(sum([op_stats.total_op_cost for _, op_stats in phys_op_stats.items()]) for _, phys_op_stats in self.operator_stats.items())

    def add_record_op_stats(self, record_op_stats: RecordOpStats | list[RecordOpStats]) -> None:
        """
        Add the given RecordOpStats to this plan's operator stats.
        """
        # normalize input type to be list[RecordOpStats]
        record_op_stats_lst = record_op_stats if isinstance(record_op_stats, list) else [record_op_stats]

        # update operator stats
        for record_op_stats in record_op_stats_lst:
            logical_op_id = record_op_stats.logical_op_id
            full_op_id = record_op_stats.full_op_id
            if logical_op_id in self.operator_stats:
                if full_op_id in self.operator_stats[logical_op_id]:
                    self.operator_stats[logical_op_id][full_op_id] += record_op_stats
                else:
                    raise ValueError(f"RecordOpStats with full_op_id {full_op_id} not found in SentinelPlanStats")
            else:
                raise ValueError(f"RecordOpStats with logical_op_id {logical_op_id} not found in SentinelPlanStats")

    def __iadd__(self, plan_stats: SentinelPlanStats) -> None:
        """
        NOTE: we assume the execution layer guarantees:
        1. these plan_stats belong to the same plan
        2. these plan_stats come from sequential (non-overlapping) executions of the same plan

        The latter criteria implies it is okay for this method to sum the plan (and operator) runtimes.
        """
        self.total_plan_time += plan_stats.total_plan_time
        self.total_plan_cost += plan_stats.total_plan_cost
        for logical_op_id, physical_op_stats in plan_stats.operator_stats.items():
            for full_op_id, op_stats in physical_op_stats.items():
                if logical_op_id in self.operator_stats:
                    if full_op_id in self.operator_stats[logical_op_id]:
                        self.operator_stats[logical_op_id][full_op_id] += op_stats
                    else:
                        self.operator_stats[logical_op_id][full_op_id] = op_stats
                else:
                    self.operator_stats[logical_op_id] = physical_op_stats

    def __str__(self) -> str:
        stats = f"total_plan_time={self.total_plan_time} \n"
        stats += f"total_plan_cost={self.total_plan_cost} \n"
        for outer_idx, physical_op_stats in enumerate(self.operator_stats.values()):
            total_time = sum([op_stats.total_op_time for op_stats in physical_op_stats.values()])
            total_cost = sum([op_stats.total_op_cost for op_stats in physical_op_stats.values()])
            stats += f"{outer_idx}. total_time={total_time} total_cost={total_cost} \n"
            for inner_idx, op_stats in enumerate(physical_op_stats.values()):
                stats += f"    {outer_idx}.{inner_idx}. {op_stats.op_name} time={op_stats.total_op_time} cost={op_stats.total_op_cost} \n"
        return stats

    def to_json(self) -> dict:
        return {
            "plan_id": self.plan_id,
            "plan_str": self.plan_str,
            "operator_stats": {
                logical_op_id: {full_op_id: op_stats.to_json() for full_op_id, op_stats in physical_op_stats.items()}
                for logical_op_id, physical_op_stats in self.operator_stats.items()
            },
            "total_plan_time": self.total_plan_time,
            "total_plan_cost": self.total_plan_cost,
        }


@dataclass
class ExecutionStats:
    """
    Dataclass for storing statistics captured for the entire execution of a workload.
    """

    # string for identifying this workload execution
    execution_id: str | None = None

    # dictionary of SentinelPlanStats objects (one for each sentinel plan run during execution)
    sentinel_plan_stats: dict[str, SentinelPlanStats] = field(default_factory=dict)

    # dictionary of PlanStats objects (one for each plan run during execution)
    plan_stats: dict[str, PlanStats] = field(default_factory=dict)

    # total time spent optimizing
    optimization_time: float = 0.0

    # total cost of optimizing
    optimization_cost: float = 0.0

    # total time spent executing the optimized plan
    plan_execution_time: float = 0.0

    # total cost of executing the optimized plan
    plan_execution_cost: float = 0.0

    # total runtime for the entire execution
    total_execution_time: float = 0.0

    # total cost for the entire execution
    total_execution_cost: float = 0.0

    # dictionary of sentinel plan strings; useful for printing executed sentinel plans in demos
    sentinel_plan_strs: dict[str, str] = field(default_factory=dict)

    # dictionary of plan strings; useful for printing executed plans in demos
    plan_strs: dict[str, str] = field(default_factory=dict)

    # start time for the execution; should be set by calling ExecutionStats.start()
    start_time: float | None = None

    # end time for the optimization; 
    optimization_end_time: float | None = None

    def start(self) -> None:
        """Start the timer for this execution."""
        self.start_time = time.time()

    def finish_optimization(self) -> None:
        """Finish the timer for the optimization phase of this execution."""
        if self.start_time is None:
            raise RuntimeError("ExecutionStats.start() must be called before ExecutionStats.finish_optimization()")

        # compute optimization time and cost
        self.optimization_end_time = time.time()
        self.optimization_time = self.optimization_end_time - self.start_time
        self.optimization_cost = self.sum_sentinel_plan_costs()

        # compute sentinel_plan_strs
        self.sentinel_plan_strs = {plan_id: plan_stats.plan_str for plan_id, plan_stats in self.sentinel_plan_stats.items()}

    def finish(self) -> None:
        """Finish the timer for this execution."""
        if self.start_time is None:
            raise RuntimeError("ExecutionStats.start() must be called before ExecutionStats.finish()")

        # compute time for plan and total execution
        end_time = time.time()
        self.plan_execution_time = (
            end_time - self.optimization_end_time
            if self.optimization_end_time is not None
            else end_time - self.start_time
        )
        self.total_execution_time = end_time - self.start_time

        # compute the cost for plan and total execution
        self.plan_execution_cost = self.sum_plan_costs()
        self.total_execution_cost = self.optimization_cost + self.plan_execution_cost

        # compute plan_strs
        self.plan_strs = {plan_id: plan_stats.plan_str for plan_id, plan_stats in self.plan_stats.items()}

    def sum_sentinel_plan_costs(self) -> float:
        """
        Sum the costs of all SentinelPlans in this execution.
        """
        return sum([plan_stats.sum_op_costs() for _, plan_stats in self.sentinel_plan_stats.items()])

    def sum_plan_costs(self) -> float:
        """
        Sum the costs of all PhysicalPlans in this execution.
        """
        return sum([plan_stats.sum_op_costs() for _, plan_stats in self.plan_stats.items()])

    def add_plan_stats(self, plan_stats: PlanStats | SentinelPlanStats | list[PlanStats] | list[SentinelPlanStats]) -> None:
        """
        Add the given PlanStats (or SentinelPlanStats) to this execution's plan stats.

        NOTE: we make the assumption that the same plan cannot be run more than once in parallel,
        i.e. each plan stats object for an individual plan comes from two different (sequential)
        periods in time. Thus, PlanStats objects can be summed.
        """
        # normalize input type to be list[PlanStats] or list[SentinelPlanStats]
        if isinstance(plan_stats, (PlanStats, SentinelPlanStats)):
            plan_stats = [plan_stats]

        for plan_stats_obj in plan_stats:
            if isinstance(plan_stats_obj, PlanStats) and plan_stats_obj.plan_id not in self.plan_stats:
                self.plan_stats[plan_stats_obj.plan_id] = plan_stats_obj
            elif isinstance(plan_stats_obj, PlanStats):
                self.plan_stats[plan_stats_obj.plan_id] += plan_stats_obj
            elif isinstance(plan_stats_obj, SentinelPlanStats) and plan_stats_obj.plan_id not in self.sentinel_plan_stats:
                self.sentinel_plan_stats[plan_stats_obj.plan_id] = plan_stats_obj
            elif isinstance(plan_stats_obj, SentinelPlanStats):
                self.sentinel_plan_stats[plan_stats_obj.plan_id] += plan_stats_obj
            else:
                raise TypeError(f"Cannot add {type(plan_stats)} to ExecutionStats")

    def clean_json(self, stats: dict):
        """
        Convert np.int64 and np.float64 to int and float for all values in stats.
        """
        for key, value in stats.items():
            if isinstance(value, dict):
                stats[key] = self.clean_json(value)
            elif isinstance(value, np.int64):
                stats[key] = int(value)
            elif isinstance(value, np.float64):
                stats[key] = float(value)
        return stats

    def to_json(self):
        stats = {
            "execution_id": self.execution_id,
            "sentinel_plan_stats": {
                plan_id: plan_stats.to_json() for plan_id, plan_stats in self.sentinel_plan_stats.items()
            },
            "plan_stats": {plan_id: plan_stats.to_json() for plan_id, plan_stats in self.plan_stats.items()},
            "optimization_time": self.optimization_time,
            "optimization_cost": self.optimization_cost,
            "plan_execution_time": self.plan_execution_time,
            "plan_execution_cost": self.plan_execution_cost,
            "total_execution_time": self.total_execution_time,
            "total_execution_cost": self.total_execution_cost,
            "sentinel_plan_strs": self.sentinel_plan_strs,
            "plan_strs": self.plan_strs,
        }
        stats = self.clean_json(stats)
        return stats


@dataclass
class OperatorCostEstimates:
    """
    Dataclass for storing estimates of key metrics of interest for each operator.
    """

    # (estimated) number of records output by this operator
    cardinality: float

    # (estimated) avg. time spent in this operator per-record
    time_per_record: float

    # (estimated) dollars spent per-record by this operator
    cost_per_record: float

    # (estimated) quality of the output from this operator
    quality: float

    # lower bound on cardinality
    cardinality_lower_bound: float | None = None

    # upper bound on cardinality
    cardinality_upper_bound: float | None = None

    # lower bound on time_per_record
    time_per_record_lower_bound: float | None = None

    # upper bound on time_per_record
    time_per_record_upper_bound: float | None = None

    # lower bound on cost_per_record
    cost_per_record_lower_bound: float | None = None

    # upper bound on cost_per_record
    cost_per_record_upper_bound: float | None = None

    # lower bound on quality
    quality_lower_bound: float | None = None

    # upper bound on quality
    quality_upper_bound: float | None = None

    def __rmul__(self, multiplier: float) -> OperatorCostEstimates:
        """
        Multiply all fields by a scalar.
        """
        dct = {field.name: getattr(self, field.name) * multiplier for field in fields(self)}
        return OperatorCostEstimates(**dct)

    def __post_init__(self):
        if self.cardinality_lower_bound is None and self.cardinality_upper_bound is None:
            self.cardinality_lower_bound = self.cardinality
            self.cardinality_upper_bound = self.cardinality

        if self.time_per_record_lower_bound is None and self.time_per_record_upper_bound is None:
            self.time_per_record_lower_bound = self.time_per_record
            self.time_per_record_upper_bound = self.time_per_record

        if self.cost_per_record_lower_bound is None and self.cost_per_record_upper_bound is None:
            self.cost_per_record_lower_bound = self.cost_per_record
            self.cost_per_record_upper_bound = self.cost_per_record

        if self.quality_lower_bound is None and self.quality_upper_bound is None:
            self.quality_lower_bound = self.quality
            self.quality_upper_bound = self.quality


@dataclass
class PlanCost:
    """
    Dataclass for storing the (cost, time, quality) estimates of (sub)-plans and their upper and lower bounds.
    """

    # the expression cost
    cost: float

    # the expression runtime
    time: float

    # the expression quality
    quality: float

    # operator-specific cost estimates
    op_estimates: OperatorCostEstimates | None = None

    # lower bound on the expression cost
    cost_lower_bound: float | None = None

    # upper bound on the expression cost
    cost_upper_bound: float | None = None

    # lower bound on the expression time
    time_lower_bound: float | None = None

    # upper bound on the expression time
    time_upper_bound: float | None = None

    # lower bound on the expression quality
    quality_lower_bound: float | None = None

    # upper bound on the expression quality
    quality_upper_bound: float | None = None

    def __hash__(self):
        return hash(f"{self.cost}-{self.time}-{self.quality}")

    def __post_init__(self):
        if self.time_lower_bound is None and self.time_upper_bound is None:
            self.time_lower_bound = self.time
            self.time_upper_bound = self.time

        if self.cost_lower_bound is None and self.cost_upper_bound is None:
            self.cost_lower_bound = self.cost
            self.cost_upper_bound = self.cost

        if self.quality_lower_bound is None and self.quality_upper_bound is None:
            self.quality_lower_bound = self.quality
            self.quality_upper_bound = self.quality

    def __iadd__(self, other: PlanCost) -> PlanCost:
        """
        NOTE: we currently assume the updating of the op_estimates are handled by the caller
        as there is not a universally correct meaning of addition of op_estiamtes.
        """
        self.cost += other.cost
        self.time += other.time
        self.quality *= other.quality
        for dataclass_field in ["cost_lower_bound", "cost_upper_bound", "time_lower_bound", "time_upper_bound"]:
            if getattr(self, dataclass_field) is not None and getattr(other, dataclass_field) is not None:
                summation = getattr(self, dataclass_field) + getattr(other, dataclass_field)
                setattr(self, dataclass_field, summation)

        for dataclass_field in ["quality_lower_bound", "quality_upper_bound"]:
            if getattr(self, dataclass_field) is not None and getattr(other, dataclass_field) is not None:
                product = getattr(self, dataclass_field) * getattr(other, dataclass_field)
                setattr(self, dataclass_field, product)

        return self

    def __add__(self, other: PlanCost) -> PlanCost:
        """
        NOTE: we currently assume the updating of the op_estimates are handled by the caller
        as there is not a universally correct meaning of addition of op_estiamtes.
        """
        dct = {
            field: getattr(self, field) + getattr(other, field)
            for field in [
                "cost",
                "cost_lower_bound",
                "cost_upper_bound",
                "time",
                "time_lower_bound",
                "time_upper_bound",
            ]
        }
        for dataclass_field in ["quality", "quality_lower_bound", "quality_upper_bound"]:
            dct[dataclass_field] = getattr(self, dataclass_field) * getattr(other, dataclass_field)

        return PlanCost(**dct)



# for those of us who resist change and are still on Python 3.9
try:
    from itertools import pairwise
except Exception:
    from more_itertools import pairwise

import gradio as gr
import pandas as pd


def flatten_nested_tuples(nested_tuples):
    """
    This function takes a nested iterable of the form (4,(3,(2,(1,())))) and flattens it to (1, 2, 3, 4).
    """
    result = []

    def flatten(item):
        if isinstance(item, tuple):
            if item:  # Check if not an empty list
                flatten(item[0])  # Process the head
                flatten(item[1])  # Process the tail
        else:
            result.append(item)

    flatten(nested_tuples)
    result = list(result)
    result.reverse()
    return result[1:]


def create_plan_str(flatten_ops):
    """Helper function to return string w/physical plan."""
    plan_str = ""
    start = flatten_ops[0]
    plan_str += f" 0. {type(start).__name__} -> {start.output_schema.__name__} \n"

    for idx, (left, right) in enumerate(pairwise(flatten_ops)):
        in_schema = left.output_schema
        out_schema = right.output_schema
        plan_str += f" {idx+1}. {in_schema.__name__} -> {type(right).__name__} -> {out_schema.__name__} "
        if hasattr(right, "model"):
            plan_str += f"\n    Using {right.model}"
            if hasattr(right, "filter"):
                filter_str = (
                    right.filter.filter_condition
                    if right.filter.filter_condition is not None
                    else str(right.filter.filter_fn)
                )
                plan_str += f'\n    Filter: "{filter_str}"'
        plan_str += "\n"
        plan_str += (
            f"    ({','.join(in_schema.field_names())[:15]}...) -> ({','.join(out_schema.field_names())[:15]}...)"
        )
        plan_str += "\n\n"

    return plan_str


def print_table(records, cols=None, plan_str=None):
    """Helper function to print execution results using Gradio"""
    if len(records) == 0:
        print("No records met search criteria")
        return

    records = [record.to_dict() for record in records]
    records_df = pd.DataFrame(records)
    print_cols = records_df.columns if cols is None else cols

    with gr.Blocks() as demo:
        gr.Dataframe(records_df[print_cols])

        if plan_str is not None:
            gr.Textbox(value=plan_str, info="Physical Plan")

    demo.launch()



from __future__ import annotations

import json
from typing import Callable

from palimpzest.constants import AggFunc, Cardinality
from palimpzest.core.data.datareaders import DataReader
from palimpzest.core.elements.filters import Filter
from palimpzest.core.elements.groupbysig import GroupBySig
from palimpzest.core.lib.schemas import Schema
from palimpzest.utils.hash_helpers import hash_for_id


class LogicalOperator:
    """
    A logical operator is an operator that operates on Sets.

    Right now it can be one of:
    - BaseScan (scans data from DataReader)
    - CacheScan (scans cached Set)
    - FilteredScan (scans input Set and applies filter)
    - ConvertScan (scans input Set and converts it to new Schema)
    - LimitScan (scans up to N records from a Set)
    - GroupByAggregate (applies a group by on the Set)
    - Aggregate (applies an aggregation on the Set)
    - RetrieveScan (fetches documents from a provided input for a given query)
    - Map (applies a function to each record in the Set without adding any new columns)

    Every logical operator must declare the get_logical_id_params() and get_logical_op_params() methods,
    which return dictionaries of parameters that are used to compute the logical op id and to implement
    the logical operator (respectively).
    """

    def __init__(
        self,
        output_schema: Schema,
        input_schema: Schema | None = None,
    ):
        self.output_schema = output_schema
        self.input_schema = input_schema
        self.logical_op_id: str | None = None

        # compute the fields generated by this logical operator
        input_field_names = self.input_schema.field_names() if self.input_schema is not None else []
        self.generated_fields = sorted(
            [field_name for field_name in self.output_schema.field_names() if field_name not in input_field_names]
        )

    def __str__(self) -> str:
        raise NotImplementedError("Abstract method")

    def __eq__(self, other) -> bool:
        all_id_params_match = all(value == getattr(other, key) for key, value in self.get_logical_id_params().items())
        return isinstance(other, self.__class__) and all_id_params_match

    def copy(self) -> LogicalOperator:
        return self.__class__(**self.get_logical_op_params())

    def logical_op_name(self) -> str:
        """Name of the logical operator."""
        return str(self.__class__.__name__)

    def get_logical_id_params(self) -> dict:
        """
        Returns a dictionary mapping of logical operator parameters which are relevant
        for computing the logical operator id.

        NOTE: Should be overriden by subclasses to include class-specific parameters.
        NOTE: input_schema and output_schema are not included in the id params because
              they depend on how the Optimizer orders operations.
        """
        return {"generated_fields": self.generated_fields}

    def get_logical_op_params(self) -> dict:
        """
        Returns a dictionary mapping of logical operator parameters which may be used to
        implement a physical operator associated with this logical operation.

        NOTE: Should be overriden by subclasses to include class-specific parameters.
        """
        return {"input_schema": self.input_schema, "output_schema": self.output_schema}

    def get_logical_op_id(self):
        """
        NOTE: We do not call this in the __init__() method as subclasses may set parameters
              returned by self.get_logical_op_params() after they call to super().__init__().
        """
        # return self.logical_op_id if we've computed it before
        if self.logical_op_id is not None:
            return self.logical_op_id

        # get op name and op parameters which are relevant for computing the id
        logical_op_name = self.logical_op_name()
        logical_id_params = self.get_logical_id_params()
        logical_id_params = {k: str(v) for k, v in logical_id_params.items()}

        # compute, set, and return the op_id
        hash_str = json.dumps({"logical_op_name": logical_op_name, **logical_id_params}, sort_keys=True)
        self.logical_op_id = hash_for_id(hash_str)

        return self.logical_op_id

    def get_generated_fields(self) -> list[str]:
        """Returns the names of the fields generated by this logical operator."""
        return self.generated_fields

    def __hash__(self):
        if not self.logical_op_id:
            raise ValueError("logical_op_id not set, unable to hash")
        return int(self.logical_op_id, 16)


class Aggregate(LogicalOperator):
    """
    Aggregate is a logical operator that applies an aggregation to the input set and yields a single result.
    This is a base class that has to be further specialized to implement specific aggregation functions.
    """

    def __init__(
        self,
        agg_func: AggFunc,
        target_cache_id: str | None = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.agg_func = agg_func
        self.target_cache_id = target_cache_id

    def __str__(self):
        return f"{self.__class__.__name__}(function: {str(self.agg_func.value)})"

    def get_logical_id_params(self) -> dict:
        logical_id_params = super().get_logical_id_params()
        logical_id_params = {"agg_func": self.agg_func, **logical_id_params}

        return logical_id_params

    def get_logical_op_params(self) -> dict:
        logical_op_params = super().get_logical_op_params()
        logical_op_params = {
            "agg_func": self.agg_func,
            "target_cache_id": self.target_cache_id,
            **logical_op_params,
        }

        return logical_op_params


class BaseScan(LogicalOperator):
    """A BaseScan is a logical operator that represents a scan of a particular data source."""

    def __init__(self, datareader: DataReader, output_schema: Schema):
        super().__init__(output_schema=output_schema)
        self.datareader = datareader

    def __str__(self):
        return f"BaseScan({self.datareader},{self.output_schema})"

    def __eq__(self, other) -> bool:
        return (
            isinstance(other, BaseScan)
            and self.input_schema.get_desc() == other.input_schema.get_desc()
            and self.output_schema.get_desc() == other.output_schema.get_desc()
            and self.datareader == other.datareader
        )

    def get_logical_id_params(self) -> dict:
        return super().get_logical_id_params()

    def get_logical_op_params(self) -> dict:
        logical_op_params = super().get_logical_op_params()
        logical_op_params = {"datareader": self.datareader, **logical_op_params}

        return logical_op_params


class CacheScan(LogicalOperator):
    """A CacheScan is a logical operator that represents a scan of a cached Set."""

    def __init__(self, datareader: DataReader, output_schema: Schema):
        super().__init__(output_schema=output_schema)
        self.datareader = datareader

    def __str__(self):
        return f"CacheScan({self.datareader},{self.output_schema})"

    def get_logical_id_params(self) -> dict:
        return super().get_logical_id_params()

    def get_logical_op_params(self) -> dict:
        logical_op_params = super().get_logical_op_params()
        logical_op_params = {"datareader": self.datareader, **logical_op_params}

        return logical_op_params


class ConvertScan(LogicalOperator):
    """A ConvertScan is a logical operator that represents a scan of a particular data source, with conversion applied."""

    def __init__(
        self,
        cardinality: Cardinality = Cardinality.ONE_TO_ONE,
        udf: Callable | None = None,
        depends_on: list[str] | None = None,
        desc: str | None = None,
        target_cache_id: str | None = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.cardinality = cardinality
        self.udf = udf
        self.depends_on = [] if depends_on is None else sorted(depends_on)
        self.desc = desc
        self.target_cache_id = target_cache_id

    def __str__(self):
        return f"ConvertScan({self.input_schema} -> {str(self.output_schema)},{str(self.desc)})"

    def get_logical_id_params(self) -> dict:
        logical_id_params = super().get_logical_id_params()
        logical_id_params = {
            "cardinality": self.cardinality,
            "udf": self.udf,
            **logical_id_params,
        }

        return logical_id_params

    def get_logical_op_params(self) -> dict:
        logical_op_params = super().get_logical_op_params()
        logical_op_params = {
            "cardinality": self.cardinality,
            "udf": self.udf,
            "depends_on": self.depends_on,
            "desc": self.desc,
            "target_cache_id": self.target_cache_id,
            **logical_op_params,
        }

        return logical_op_params


class FilteredScan(LogicalOperator):
    """A FilteredScan is a logical operator that represents a scan of a particular data source, with filters applied."""

    def __init__(
        self,
        filter: Filter,
        depends_on: list[str] | None = None,
        target_cache_id: str | None = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.filter = filter
        self.depends_on = [] if depends_on is None else sorted(depends_on)
        self.target_cache_id = target_cache_id

    def __str__(self):
        return f"FilteredScan({str(self.output_schema)}, {str(self.filter)})"

    def get_logical_id_params(self) -> dict:
        logical_id_params = super().get_logical_id_params()
        logical_id_params = {
            "filter": self.filter,
            **logical_id_params,
        }

        return logical_id_params

    def get_logical_op_params(self) -> dict:
        logical_op_params = super().get_logical_op_params()
        logical_op_params = {
            "filter": self.filter,
            "depends_on": self.depends_on,
            "target_cache_id": self.target_cache_id,
            **logical_op_params,
        }

        return logical_op_params


class GroupByAggregate(LogicalOperator):
    def __init__(
        self,
        group_by_sig: GroupBySig,
        target_cache_id: str | None = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        if not self.input_schema:
            raise ValueError("GroupByAggregate requires an input schema")
        (valid, error) = group_by_sig.validate_schema(self.input_schema)
        if not valid:
            raise TypeError(error)
        self.group_by_sig = group_by_sig
        self.target_cache_id = target_cache_id

    def __str__(self):
        return f"GroupBy({self.group_by_sig.serialize()})"

    def get_logical_id_params(self) -> dict:
        logical_id_params = super().get_logical_id_params()
        logical_id_params = {"group_by_sig": self.group_by_sig, **logical_id_params}

        return logical_id_params

    def get_logical_op_params(self) -> dict:
        logical_op_params = super().get_logical_op_params()
        logical_op_params = {
            "group_by_sig": self.group_by_sig,
            "target_cache_id": self.target_cache_id,
            **logical_op_params,
        }

        return logical_op_params


class LimitScan(LogicalOperator):
    def __init__(self, limit: int, target_cache_id: str | None = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.limit = limit
        self.target_cache_id = target_cache_id

    def __str__(self):
        return f"LimitScan({str(self.input_schema)}, {str(self.output_schema)})"

    def get_logical_id_params(self) -> dict:
        logical_id_params = super().get_logical_id_params()
        logical_id_params = {"limit": self.limit, **logical_id_params}

        return logical_id_params

    def get_logical_op_params(self) -> dict:
        logical_op_params = super().get_logical_op_params()
        logical_op_params = {
            "limit": self.limit,
            "target_cache_id": self.target_cache_id,
            **logical_op_params,
        }

        return logical_op_params


class Project(LogicalOperator):
    def __init__(self, project_cols: list[str], target_cache_id: str | None = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.project_cols = project_cols
        self.target_cache_id = target_cache_id

    def __str__(self):
        return f"Project({self.input_schema}, {self.project_cols})"

    def get_logical_id_params(self) -> dict:
        logical_id_params = super().get_logical_id_params()
        logical_id_params = {"project_cols": self.project_cols, **logical_id_params}

        return logical_id_params

    def get_logical_op_params(self) -> dict:
        logical_op_params = super().get_logical_op_params()
        logical_op_params = {
            "project_cols": self.project_cols,
            "target_cache_id": self.target_cache_id,
            **logical_op_params,
        }

        return logical_op_params


class RetrieveScan(LogicalOperator):
    """A RetrieveScan is a logical operator that represents a scan of a particular data source, with a convert-like retrieve applied."""

    def __init__(
        self,
        index,
        search_func,
        search_attr,
        output_attrs,
        k,
        target_cache_id: str = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.index = index
        self.search_func = search_func
        self.search_attr = search_attr
        self.output_attrs = output_attrs
        self.k = k
        self.target_cache_id = target_cache_id

    def __str__(self):
        return f"RetrieveScan({self.input_schema} -> {str(self.output_schema)},{str(self.desc)})"

    def get_logical_id_params(self) -> dict:
        # NOTE: if we allow optimization over index, then we will need to include it in the id params
        # NOTE: ^if we do this, we should probably make a wrapper around the index object to ensure that
        #       it can be serialized as a string properly
        logical_id_params = super().get_logical_id_params()
        logical_id_params = {
            "search_attr": self.search_attr,
            "output_attrs": self.output_attrs,
            "k": self.k,
            **logical_id_params,
        }

        return logical_id_params

    def get_logical_op_params(self) -> dict:
        logical_op_params = super().get_logical_op_params()
        logical_op_params = {
            "index": self.index,
            "search_func": self.search_func,
            "search_attr": self.search_attr,
            "output_attrs": self.output_attrs,
            "k": self.k,
            "target_cache_id": self.target_cache_id,
            **logical_op_params,
        }

        return logical_op_params


# TODO: (near-term) maybe we should try to fold this into ConvertScan, and make the internals of PZ
#       amenable to a convert operator (with a UDF) that does not add new columns?
class MapScan(LogicalOperator):
    """A MapScan is a logical operator that applies a UDF to each input record without adding new columns."""

    def __init__(
        self,
        udf: Callable | None = None,
        target_cache_id: str | None = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.udf = udf
        self.target_cache_id = target_cache_id

    def __str__(self):
        return f"MapScan({self.output_schema}, {self.udf.__name__})"

    def get_logical_id_params(self) -> dict:
        logical_id_params = super().get_logical_id_params()
        logical_id_params = {
            "udf": self.udf,
            **logical_id_params,
        }

        return logical_id_params

    def get_logical_op_params(self) -> dict:
        logical_op_params = super().get_logical_op_params()
        logical_op_params = {
            "udf": self.udf,
            "target_cache_id": self.target_cache_id,
            **logical_op_params,
        }

        return logical_op_params
